import pytest
import salt.states.proxy as proxy
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {proxy: {}}


def test_set_proxy_macos():
    """
    Test to make sure we can set the proxy settings on macOS
    """
    with patch.dict(proxy.__grains__, {"os": "Darwin"}):
        expected = {
            "changes": {
                "new": [
                    {
                        "port": "3128",
                        "server": "192.168.0.1",
                        "service": "http",
                        "user": "frank",
                    },
                    {
                        "port": "3128",
                        "server": "192.168.0.1",
                        "service": "https",
                        "user": "frank",
                    },
                    {
                        "port": "3128",
                        "server": "192.168.0.1",
                        "service": "ftp",
                        "user": "frank",
                    },
                    {"bypass_domains": ["salt.com", "test.com"]},
                ]
            },
            "comment": (
                "http proxy settings updated correctly\nhttps "
                "proxy settings updated correctly\nftp proxy "
                "settings updated correctly\nProxy "
                "bypass domains updated correctly\n"
            ),
            "name": "192.168.0.1",
            "result": True,
        }

        set_proxy_mock = MagicMock(return_value=True)
        patches = {
            "proxy.get_http_proxy": MagicMock(return_value={}),
            "proxy.get_https_proxy": MagicMock(return_value={}),
            "proxy.get_ftp_proxy": MagicMock(return_value={}),
            "proxy.get_proxy_bypass": MagicMock(return_value=[]),
            "proxy.set_http_proxy": set_proxy_mock,
            "proxy.set_https_proxy": set_proxy_mock,
            "proxy.set_ftp_proxy": set_proxy_mock,
            "proxy.set_proxy_bypass": set_proxy_mock,
        }

        with patch.dict(proxy.__salt__, patches):
            out = proxy.managed(
                "192.168.0.1",
                "3128",
                user="frank",
                password="passw0rd",
                bypass_domains=["salt.com", "test.com"],
            )
            out["changes"]["new"][-1]["bypass_domains"] = sorted(
                out["changes"]["new"][-1]["bypass_domains"]
            )

            calls = [
                call("192.168.0.1", "3128", "frank", "passw0rd", "Ethernet"),
                call("192.168.0.1", "3128", "frank", "passw0rd", "Ethernet"),
                call("192.168.0.1", "3128", "frank", "passw0rd", "Ethernet"),
                call(["salt.com", "test.com"], "Ethernet"),
            ]

            set_proxy_mock.assert_has_calls(calls)
            assert out == expected


def test_set_proxy_macos_same():
    """
    Test to make sure we can set the proxy settings on macOS
    """
    with patch.dict(proxy.__grains__, {"os": "Darwin"}):
        expected = {
            "changes": {},
            "comment": (
                "http proxy settings already set.\nhttps "
                "proxy settings already set.\nftp proxy settings"
                " already set.\nProxy bypass domains "
                "are already set correctly.\n"
            ),
            "name": "192.168.0.1",
            "result": True,
        }

        proxy_val = {"enabled": True, "server": "192.168.0.1", "port": "3128"}

        set_proxy_mock = MagicMock()
        patches = {
            "proxy.get_http_proxy": MagicMock(return_value=proxy_val),
            "proxy.get_https_proxy": MagicMock(return_value=proxy_val),
            "proxy.get_ftp_proxy": MagicMock(return_value=proxy_val),
            "proxy.get_proxy_bypass": MagicMock(return_value=["test.com", "salt.com"]),
            "proxy.set_http_proxy": set_proxy_mock,
            "proxy.set_https_proxy": set_proxy_mock,
            "proxy.set_ftp_proxy": set_proxy_mock,
            "proxy.set_proxy_bypass": set_proxy_mock,
        }

        with patch.dict(proxy.__salt__, patches):
            out = proxy.managed(
                "192.168.0.1",
                "3128",
                user="frank",
                password="passw0rd",
                bypass_domains=["salt.com", "test.com"],
            )

            assert not set_proxy_mock.called
            assert out == expected


def test_set_proxy_windows():
    """
    Test to make sure we can set the proxy settings on Windows
    """
    with patch.dict(proxy.__grains__, {"os": "Windows"}):
        expected = {
            "changes": {},
            "comment": "Proxy settings updated correctly",
            "name": "192.168.0.1",
            "result": True,
        }

        set_proxy_mock = MagicMock(return_value=True)
        patches = {
            "proxy.get_proxy_win": MagicMock(return_value={}),
            "proxy.get_proxy_bypass": MagicMock(return_value=[]),
            "proxy.set_proxy_win": set_proxy_mock,
        }

        with patch.dict(proxy.__salt__, patches):
            out = proxy.managed(
                "192.168.0.1",
                "3128",
                user="frank",
                password="passw0rd",
                bypass_domains=["salt.com", "test.com"],
            )

            set_proxy_mock.assert_called_once_with(
                "192.168.0.1",
                "3128",
                ["http", "https", "ftp"],
                ["salt.com", "test.com"],
            )
            assert out == expected


def test_set_proxy_windows_same():
    """
    Test to make sure we can set the proxy settings on Windows
    """
    with patch.dict(proxy.__grains__, {"os": "Windows"}):
        expected = {
            "changes": {},
            "comment": "Proxy settings already correct.",
            "name": "192.168.0.1",
            "result": True,
        }

        proxy_val = {
            "enabled": True,
            "http": {"enabled": True, "server": "192.168.0.1", "port": "3128"},
            "https": {"enabled": True, "server": "192.168.0.1", "port": "3128"},
            "ftp": {"enabled": True, "server": "192.168.0.1", "port": "3128"},
        }

        set_proxy_mock = MagicMock(return_value=True)
        patches = {
            "proxy.get_proxy_win": MagicMock(return_value=proxy_val),
            "proxy.get_proxy_bypass": MagicMock(return_value=["salt.com", "test.com"]),
            "proxy.set_proxy_win": set_proxy_mock,
        }

        with patch.dict(proxy.__salt__, patches):
            out = proxy.managed(
                "192.168.0.1",
                "3128",
                user="frank",
                password="passw0rd",
                bypass_domains=["salt.com", "test.com"],
            )

            assert not set_proxy_mock.called
            assert out == expected



"""
Functions for interacting with the job cache
"""


import logging

import salt.minion
import salt.utils.event
import salt.utils.jid
import salt.utils.verify

log = logging.getLogger(__name__)


def store_job(opts, load, event=None, mminion=None):
    """
    Store job information using the configured master_job_cache
    """
    # Generate EndTime
    endtime = salt.utils.jid.jid_to_time(salt.utils.jid.gen_jid(opts))
    # If the return data is invalid, just ignore it
    if any(key not in load for key in ("return", "jid", "id")):
        return False
    if not salt.utils.verify.valid_id(opts, load["id"]):
        return False
    if mminion is None:
        mminion = salt.minion.MasterMinion(opts, states=False, rend=False)

    job_cache = opts["master_job_cache"]
    if load["jid"] == "req":
        # The minion is returning a standalone job, request a jobid
        load["arg"] = load.get("arg", load.get("fun_args", []))
        load["tgt_type"] = "glob"
        load["tgt"] = load["id"]

        prep_fstr = "{}.prep_jid".format(opts["master_job_cache"])
        try:
            load["jid"] = mminion.returners[prep_fstr](
                nocache=load.get("nocache", False)
            )
        except KeyError:
            emsg = "Returner '{}' does not support function prep_jid".format(job_cache)
            log.error(emsg)
            raise KeyError(emsg)
        except Exception:  # pylint: disable=broad-except
            log.critical(
                "The specified '%s' returner threw a stack trace:\n",
                job_cache,
                exc_info=True,
            )

        # save the load, since we don't have it
        saveload_fstr = "{}.save_load".format(job_cache)
        try:
            mminion.returners[saveload_fstr](load["jid"], load)
        except KeyError:
            emsg = "Returner '{}' does not support function save_load".format(job_cache)
            log.error(emsg)
            raise KeyError(emsg)
        except Exception:  # pylint: disable=broad-except
            log.critical(
                "The specified '%s' returner threw a stack trace",
                job_cache,
                exc_info=True,
            )
    elif salt.utils.jid.is_jid(load["jid"]):
        # Store the jid
        jidstore_fstr = "{}.prep_jid".format(job_cache)
        try:
            mminion.returners[jidstore_fstr](False, passed_jid=load["jid"])
        except KeyError:
            emsg = "Returner '{}' does not support function prep_jid".format(job_cache)
            log.error(emsg)
            raise KeyError(emsg)
        except Exception:  # pylint: disable=broad-except
            log.critical(
                "The specified '%s' returner threw a stack trace",
                job_cache,
                exc_info=True,
            )

    if event:
        # If the return data is invalid, just ignore it
        log.info("Got return from %s for job %s", load["id"], load["jid"])
        event.fire_event(
            load, salt.utils.event.tagify([load["jid"], "ret", load["id"]], "job")
        )
        event.fire_ret_load(load)

    # if you have a job_cache, or an ext_job_cache, don't write to
    # the regular master cache
    if not opts["job_cache"] or opts.get("ext_job_cache"):
        return

    # do not cache job results if explicitly requested
    if load.get("jid") == "nocache":
        log.debug(
            "Ignoring job return with jid for caching %s from %s",
            load["jid"],
            load["id"],
        )
        return

    # otherwise, write to the master cache
    savefstr = "{}.save_load".format(job_cache)
    getfstr = "{}.get_load".format(job_cache)
    fstr = "{}.returner".format(job_cache)
    updateetfstr = "{}.update_endtime".format(job_cache)
    if "fun" not in load and load.get("return", {}):
        ret_ = load.get("return", {})
        if "fun" in ret_:
            load.update({"fun": ret_["fun"]})
        if "user" in ret_:
            load.update({"user": ret_["user"]})

    # Try to reach returner methods
    try:
        savefstr_func = mminion.returners[savefstr]
        getfstr_func = mminion.returners[getfstr]
        fstr_func = mminion.returners[fstr]
    except KeyError as error:
        emsg = "Returner '{}' does not support function {}".format(job_cache, error)
        log.error(emsg)
        raise KeyError(emsg)

    if job_cache != "local_cache":
        try:
            mminion.returners[savefstr](load["jid"], load)
        except KeyError as e:
            log.error("Load does not contain 'jid': %s", e)
        except Exception:  # pylint: disable=broad-except
            log.critical(
                "The specified '%s' returner threw a stack trace",
                job_cache,
                exc_info=True,
            )

    try:
        mminion.returners[fstr](load)
    except Exception:  # pylint: disable=broad-except
        log.critical(
            "The specified '%s' returner threw a stack trace", job_cache, exc_info=True
        )

    if opts.get("job_cache_store_endtime") and updateetfstr in mminion.returners:
        mminion.returners[updateetfstr](load["jid"], endtime)


def store_minions(opts, jid, minions, mminion=None, syndic_id=None):
    """
    Store additional minions matched on lower-level masters using the configured
    master_job_cache
    """
    if mminion is None:
        mminion = salt.minion.MasterMinion(opts, states=False, rend=False)
    job_cache = opts["master_job_cache"]
    minions_fstr = "{}.save_minions".format(job_cache)

    try:
        mminion.returners[minions_fstr](jid, minions, syndic_id=syndic_id)
    except KeyError:
        raise KeyError(
            "Returner '{}' does not support function save_minions".format(job_cache)
        )


def get_retcode(ret):
    """
    Determine a retcode for a given return
    """
    retcode = 0
    # if there is a dict with retcode, use that
    if isinstance(ret, dict) and ret.get("retcode", 0) != 0:
        return ret["retcode"]
    # if its a boolean, False means 1
    elif isinstance(ret, bool) and not ret:
        return 1
    return retcode


# vim:set et sts=4 ts=4 tw=80:



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import salt.modules.nfs3 as nfs3
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class NfsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.nfs3
    """

    def setup_loader_modules(self):
        return {nfs3: {}}

    def test_list_exports(self):
        """
        Test for List configured exports
        """
        with patch("salt.utils.files.fopen", mock_open(read_data="A B1(23")):
            exports = nfs3.list_exports()
            assert exports == {"A": [{"hosts": "B1", "options": ["23"]}]}, exports

    def test_del_export(self):
        """
        Test for Remove an export
        """
        list_exports_mock = MagicMock(
            return_value={"A": [{"hosts": ["B1"], "options": ["23"]}]}
        )
        with patch.object(nfs3, "list_exports", list_exports_mock), patch.object(
            nfs3, "_write_exports", MagicMock(return_value=None)
        ):
            result = nfs3.del_export(path="A")
            assert result == {}, result



"""
Module for running ZFS command

:codeauthor:    Nitin Madhok <nmadhok@g.clemson.edu>, Jorge Schrauwen <sjorge@blackdot.be>
:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.zfs
:platform:      illumos,freebsd,linux

.. versionchanged:: 2018.3.1
  Big refactor to remove duplicate code, better type conversions and improved
  consistency in output.

"""

import logging

import salt.modules.cmdmod
import salt.utils.args
import salt.utils.path
import salt.utils.platform
import salt.utils.versions
from salt.utils.odict import OrderedDict

__virtualname__ = "zfs"
log = logging.getLogger(__name__)

# Function alias to set mapping.
__func_alias__ = {
    "list_": "list",
}


def __virtual__():
    """
    Only load when the platform has zfs support
    """
    if __grains__.get("zfs_support"):
        return __virtualname__
    else:
        return False, "The zfs module cannot be loaded: zfs not supported"


def exists(name, **kwargs):
    """
    Check if a ZFS filesystem or volume or snapshot exists.

    name : string
        name of dataset
    type : string
        also check if dataset is of a certain type, valid choices are:
        filesystem, snapshot, volume, bookmark, or all.

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.exists myzpool/mydataset
        salt '*' zfs.exists myzpool/myvolume type=volume

    """
    ## Configure command
    # NOTE: initialize the defaults
    opts = {}

    # NOTE: set extra config from kwargs
    if kwargs.get("type", False):
        opts["-t"] = kwargs.get("type")

    ## Check if 'name' of 'type' exists
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="list",
            opts=opts,
            target=name,
        ),
        python_shell=False,
        ignore_retcode=True,
    )

    return res["retcode"] == 0


def create(name, **kwargs):
    """
    Create a ZFS File System.

    name : string
        name of dataset or volume
    volume_size : string
        if specified, a zvol will be created instead of a dataset
    sparse : boolean
        create sparse volume
    create_parent : boolean
        creates all the non-existing parent datasets. any property specified on the
        command line using the -o option is ignored.
    properties : dict
        additional zfs properties (-o)

    .. note::

        ZFS properties can be specified at the time of creation of the filesystem by
        passing an additional argument called "properties" and specifying the properties
        with their respective values in the form of a python dictionary::

            properties="{'property1': 'value1', 'property2': 'value2'}"

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.create myzpool/mydataset [create_parent=True|False]
        salt '*' zfs.create myzpool/mydataset properties="{'mountpoint': '/export/zfs', 'sharenfs': 'on'}"
        salt '*' zfs.create myzpool/volume volume_size=1G [sparse=True|False]`
        salt '*' zfs.create myzpool/volume volume_size=1G properties="{'volblocksize': '512'}" [sparse=True|False]

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []
    opts = {}

    # NOTE: push filesystem properties
    filesystem_properties = kwargs.get("properties", {})

    # NOTE: set extra config from kwargs
    if kwargs.get("create_parent", False):
        flags.append("-p")
    if kwargs.get("sparse", False) and kwargs.get("volume_size", None):
        flags.append("-s")
    if kwargs.get("volume_size", None):
        opts["-V"] = __utils__["zfs.to_size"](
            kwargs.get("volume_size"), convert_to_human=False
        )

    ## Create filesystem/volume
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="create",
            flags=flags,
            opts=opts,
            filesystem_properties=filesystem_properties,
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "created")


def destroy(name, **kwargs):
    """
    Destroy a ZFS File System.

    name : string
        name of dataset, volume, or snapshot
    force : boolean
        force an unmount of any file systems using the unmount -f command.
    recursive : boolean
        recursively destroy all children. (-r)
    recursive_all : boolean
        recursively destroy all dependents, including cloned file systems
        outside the target hierarchy. (-R)

    .. warning::
        watch out when using recursive and recursive_all

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.destroy myzpool/mydataset [force=True|False]

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []

    # NOTE: set extra config from kwargs
    if kwargs.get("force", False):
        flags.append("-f")
    if kwargs.get("recursive_all", False):
        flags.append("-R")
    if kwargs.get("recursive", False):
        flags.append("-r")

    ## Destroy filesystem/volume/snapshot/...
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="destroy",
            flags=flags,
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "destroyed")


def rename(name, new_name, **kwargs):
    """
    Rename or Relocate a ZFS File System.

    name : string
        name of dataset, volume, or snapshot
    new_name : string
        new name of dataset, volume, or snapshot
    force : boolean
        force unmount any filesystems that need to be unmounted in the process.
    create_parent : boolean
        creates all the nonexistent parent datasets. Datasets created in
        this manner are automatically mounted according to the mountpoint
        property inherited from their parent.
    recursive : boolean
        recursively rename the snapshots of all descendent datasets.
        snapshots are the only dataset that can be renamed recursively.

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.rename myzpool/mydataset myzpool/renameddataset

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []
    target = []

    # NOTE: set extra config from kwargs
    if __utils__["zfs.is_snapshot"](name):
        if kwargs.get("create_parent", False):
            log.warning(
                "zfs.rename - create_parent=True cannot be used with snapshots."
            )
        if kwargs.get("force", False):
            log.warning("zfs.rename - force=True cannot be used with snapshots.")
        if kwargs.get("recursive", False):
            flags.append("-r")
    else:
        if kwargs.get("create_parent", False):
            flags.append("-p")
        if kwargs.get("force", False):
            flags.append("-f")
        if kwargs.get("recursive", False):
            log.warning("zfs.rename - recursive=True can only be used with snapshots.")

    # NOTE: update target
    target.append(name)
    target.append(new_name)

    ## Rename filesystem/volume/snapshot/...
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="rename",
            flags=flags,
            target=target,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "renamed")


def list_(name=None, **kwargs):
    """
    Return a list of all datasets or a specified dataset on the system and the
    values of their used, available, referenced, and mountpoint properties.

    name : string
        name of dataset, volume, or snapshot
    recursive : boolean
        recursively list children
    depth : int
        limit recursion to depth
    properties : string
        comma-separated list of properties to list, the name property will always be added
    type : string
        comma-separated list of types to display, where type is one of
        filesystem, snapshot, volume, bookmark, or all.
    sort : string
        property to sort on (default = name)
    order : string [ascending|descending]
        sort order (default = ascending)
    parsable : boolean
        display numbers in parsable (exact) values
        .. versionadded:: 2018.3.0

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.list
        salt '*' zfs.list myzpool/mydataset [recursive=True|False]
        salt '*' zfs.list myzpool/mydataset properties="sharenfs,mountpoint"

    """
    ret = OrderedDict()

    ## update properties
    # NOTE: properties should be a list
    properties = kwargs.get("properties", "used,avail,refer,mountpoint")
    if not isinstance(properties, list):
        properties = properties.split(",")

    # NOTE: name should be first property
    #       we loop here because there 'name' can be in the list
    #       multiple times.
    while "name" in properties:
        properties.remove("name")
    properties.insert(0, "name")

    ## Configure command
    # NOTE: initialize the defaults
    flags = ["-H"]
    opts = {}

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive", False):
        flags.append("-r")
    if kwargs.get("recursive", False) and kwargs.get("depth", False):
        opts["-d"] = kwargs.get("depth")
    if kwargs.get("type", False):
        opts["-t"] = kwargs.get("type")
    kwargs_sort = kwargs.get("sort", False)
    if kwargs_sort and kwargs_sort in properties:
        if kwargs.get("order", "ascending").startswith("a"):
            opts["-s"] = kwargs_sort
        else:
            opts["-S"] = kwargs_sort
    if isinstance(properties, list):
        # NOTE: There can be only one -o and it takes a comma-separated list
        opts["-o"] = ",".join(properties)
    else:
        opts["-o"] = properties

    ## parse zfs list
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="list",
            flags=flags,
            opts=opts,
            target=name,
        ),
        python_shell=False,
    )
    if res["retcode"] == 0:
        for ds in res["stdout"].splitlines():
            if kwargs.get("parsable", True):
                ds_data = __utils__["zfs.from_auto_dict"](
                    OrderedDict(list(zip(properties, ds.split("\t")))),
                )
            else:
                ds_data = __utils__["zfs.to_auto_dict"](
                    OrderedDict(list(zip(properties, ds.split("\t")))),
                    convert_to_human=True,
                )

            ret[ds_data["name"]] = ds_data
            del ret[ds_data["name"]]["name"]
    else:
        return __utils__["zfs.parse_command_result"](res)

    return ret


def list_mount():
    """
    List mounted zfs filesystems

    .. versionadded:: 2018.3.1

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.list_mount

    """
    ## List mounted filesystem
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="mount",
        ),
        python_shell=False,
    )

    if res["retcode"] == 0:
        ret = OrderedDict()
        for mount in res["stdout"].splitlines():
            mount = mount.split()
            ret[mount[0]] = mount[-1]
        return ret
    else:
        return __utils__["zfs.parse_command_result"](res)


def mount(name=None, **kwargs):
    """
    Mounts ZFS file systems

    name : string
        name of the filesystem, having this set to None will mount all filesystems. (this is the default)
    overlay : boolean
        perform an overlay mount.
    options : string
        optional comma-separated list of mount options to use temporarily for
        the duration of the mount.

    .. versionadded:: 2016.3.0
    .. versionchanged:: 2018.3.1

    .. warning::

            Passing '-a' as name is deprecated and will be removed in 3001.

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.mount
        salt '*' zfs.mount myzpool/mydataset
        salt '*' zfs.mount myzpool/mydataset options=ro

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []
    opts = {}

    # NOTE: set extra config from kwargs
    if kwargs.get("overlay", False):
        flags.append("-O")
    if kwargs.get("options", False):
        opts["-o"] = kwargs.get("options")

    ## Mount filesystem
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="mount",
            flags=flags,
            opts=opts,
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "mounted")


def unmount(name, **kwargs):
    """
    Unmounts ZFS file systems

    name : string
        name of the filesystem, you can use None to unmount all mounted filesystems.
    force : boolean
        forcefully unmount the file system, even if it is currently in use.

    .. warning::

        Using ``-a`` for the name parameter will probably break your system, unless your rootfs is not on zfs.

    .. versionadded:: 2016.3.0
    .. versionchanged:: 2018.3.1

    .. warning::

            Passing '-a' as name is deprecated and will be removed in 3001.

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.unmount myzpool/mydataset [force=True|False]

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []

    # NOTE: set extra config from kwargs
    if kwargs.get("force", False):
        flags.append("-f")
    if name in [None, "-a"]:
        # NOTE: still accept '-a' as name for backwards compatibility
        #       until Salt 3001 this should just simplify
        #       this to just set '-a' if name is not set.
        flags.append("-a")
        name = None

    ## Unmount filesystem
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="unmount",
            flags=flags,
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "unmounted")


def inherit(prop, name, **kwargs):
    """
    Clears the specified property

    prop : string
        name of property
    name : string
        name of the filesystem, volume, or snapshot
    recursive : boolean
        recursively inherit the given property for all children.
    revert : boolean
        revert the property to the received value if one exists; otherwise
        operate as if the -S option was not specified.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.inherit canmount myzpool/mydataset [recursive=True|False]

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive", False):
        flags.append("-r")
    if kwargs.get("revert", False):
        flags.append("-S")

    ## Inherit property
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="inherit",
            flags=flags,
            property_name=prop,
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "inherited")


def diff(name_a, name_b=None, **kwargs):
    """
    Display the difference between a snapshot of a given filesystem and
    another snapshot of that filesystem from a later time or the current
    contents of the filesystem.

    name_a : string
        name of snapshot
    name_b : string
        (optional) name of snapshot or filesystem
    show_changetime : boolean
        display the path's inode change time as the first column of output. (default = True)
    show_indication : boolean
        display an indication of the type of file. (default = True)
    parsable : boolean
        if true we don't parse the timestamp to a more readable date (default = True)

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.diff myzpool/mydataset@yesterday myzpool/mydataset

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = ["-H"]
    target = []

    # NOTE: set extra config from kwargs
    if kwargs.get("show_changetime", True):
        flags.append("-t")
    if kwargs.get("show_indication", True):
        flags.append("-F")

    # NOTE: update target
    target.append(name_a)
    if name_b:
        target.append(name_b)

    ## Diff filesystem/snapshot
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="diff",
            flags=flags,
            target=target,
        ),
        python_shell=False,
    )

    if res["retcode"] != 0:
        return __utils__["zfs.parse_command_result"](res)
    else:
        if not kwargs.get("parsable", True) and kwargs.get("show_changetime", True):
            ret = OrderedDict()
            for entry in res["stdout"].splitlines():
                entry = entry.split()
                entry_timestamp = __utils__["dateutils.strftime"](
                    entry[0], "%Y-%m-%d.%H:%M:%S.%f"
                )
                entry_data = "\t\t".join(entry[1:])
                ret[entry_timestamp] = entry_data
        else:
            ret = res["stdout"].splitlines()
        return ret


def rollback(name, **kwargs):
    """
    Roll back the given dataset to a previous snapshot.

    name : string
        name of snapshot
    recursive : boolean
        destroy any snapshots and bookmarks more recent than the one
        specified.
    recursive_all : boolean
        destroy any more recent snapshots and bookmarks, as well as any
        clones of those snapshots.
    force : boolean
        used with the -R option to force an unmount of any clone file
        systems that are to be destroyed.

    .. warning::

        When a dataset is rolled back, all data that has changed since
        the snapshot is discarded, and the dataset reverts to the state
        at the time of the snapshot. By default, the command refuses to
        roll back to a snapshot other than the most recent one.

        In order to do so, all intermediate snapshots and bookmarks
        must be destroyed by specifying the -r option.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.rollback myzpool/mydataset@yesterday

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive_all", False):
        flags.append("-R")
    if kwargs.get("recursive", False):
        flags.append("-r")
    if kwargs.get("force", False):
        if kwargs.get("recursive_all", False) or kwargs.get("recursive", False):
            flags.append("-f")
        else:
            log.warning(
                "zfs.rollback - force=True can only be used with recursive_all=True or"
                " recursive=True"
            )

    ## Rollback to snapshot
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="rollback",
            flags=flags,
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "rolledback")


def clone(name_a, name_b, **kwargs):
    """
    Creates a clone of the given snapshot.

    name_a : string
        name of snapshot
    name_b : string
        name of filesystem or volume
    create_parent : boolean
        creates all the non-existing parent datasets. any property specified on the
        command line using the -o option is ignored.
    properties : dict
        additional zfs properties (-o)

    .. note::

        ZFS properties can be specified at the time of creation of the filesystem by
        passing an additional argument called "properties" and specifying the properties
        with their respective values in the form of a python dictionary::

            properties="{'property1': 'value1', 'property2': 'value2'}"

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.clone myzpool/mydataset@yesterday myzpool/mydataset_yesterday

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []
    target = []

    # NOTE: push filesystem properties
    filesystem_properties = kwargs.get("properties", {})

    # NOTE: set extra config from kwargs
    if kwargs.get("create_parent", False):
        flags.append("-p")

    # NOTE: update target
    target.append(name_a)
    target.append(name_b)

    ## Clone filesystem/volume
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="clone",
            flags=flags,
            filesystem_properties=filesystem_properties,
            target=target,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "cloned")


def promote(name):
    """
    Promotes a clone file system to no longer be dependent on its "origin"
    snapshot.

    .. note::

        This makes it possible to destroy the file system that the
        clone was created from. The clone parent-child dependency relationship
        is reversed, so that the origin file system becomes a clone of the
        specified file system.

        The snapshot that was cloned, and any snapshots previous to this
        snapshot, are now owned by the promoted clone. The space they use moves
        from the origin file system to the promoted clone, so enough space must
        be available to accommodate these snapshots. No new space is consumed
        by this operation, but the space accounting is adjusted. The promoted
        clone must not have any conflicting snapshot names of its own. The
        rename subcommand can be used to rename any conflicting snapshots.

    name : string
        name of clone-filesystem

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.promote myzpool/myclone

    """
    ## Promote clone
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="promote",
            target=name,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "promoted")


def bookmark(snapshot, bookmark):
    """
    Creates a bookmark of the given snapshot

    .. note::

        Bookmarks mark the point in time when the snapshot was created,
        and can be used as the incremental source for a zfs send command.

        This feature must be enabled to be used. See zpool-features(5) for
        details on ZFS feature flags and the bookmarks feature.

    snapshot : string
        name of snapshot to bookmark
    bookmark : string
        name of bookmark

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.bookmark myzpool/mydataset@yesterday myzpool/mydataset#complete

    """
    # abort if we do not have feature flags
    if not __utils__["zfs.has_feature_flags"]():
        return OrderedDict([("error", "bookmarks are not supported")])

    ## Configure command
    # NOTE: initialize the defaults
    target = []

    # NOTE: update target
    target.append(snapshot)
    target.append(bookmark)

    ## Bookmark snapshot
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="bookmark",
            target=target,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "bookmarked")


def holds(snapshot, **kwargs):
    """
    Lists all existing user references for the given snapshot or snapshots.

    snapshot : string
        name of snapshot
    recursive : boolean
        lists the holds that are set on the named descendent snapshots also.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.holds myzpool/mydataset@baseline

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = ["-H"]
    target = []

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive", False):
        flags.append("-r")

    # NOTE: update target
    target.append(snapshot)

    ## Lookup holds
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="holds",
            flags=flags,
            target=target,
        ),
        python_shell=False,
    )

    ret = __utils__["zfs.parse_command_result"](res)
    if res["retcode"] == 0:
        for hold in res["stdout"].splitlines():
            hold_data = OrderedDict(
                list(
                    zip(
                        ["name", "tag", "timestamp"],
                        hold.split("\t"),
                    )
                )
            )
            ret[hold_data["tag"].strip()] = hold_data["timestamp"]

    return ret


def hold(tag, *snapshot, **kwargs):
    """
    Adds a single reference, named with the tag argument, to the specified
    snapshot or snapshots.

    .. note::

        Each snapshot has its own tag namespace, and tags must be unique within that space.

        If a hold exists on a snapshot, attempts to destroy that snapshot by
        using the zfs destroy command return EBUSY.

    tag : string
        name of tag
    snapshot : string
        name of snapshot(s)
    recursive : boolean
        specifies that a hold with the given tag is applied recursively to
        the snapshots of all descendent file systems.

    .. versionadded:: 2016.3.0
    .. versionchanged:: 2018.3.1

    .. warning::

        As of 2018.3.1 the tag parameter no longer accepts a comma-separated value.
        It's is now possible to create a tag that contains a comma, this was impossible before.

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.hold mytag myzpool/mydataset@mysnapshot [recursive=True]
        salt '*' zfs.hold mytag myzpool/mydataset@mysnapshot myzpool/mydataset@myothersnapshot
    """

    ## Configure command
    # NOTE: initialize the defaults
    flags = []
    target = []

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive", False):
        flags.append("-r")

    # NOTE: update target
    target.append(tag)
    target.extend(snapshot)

    ## hold snapshot
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="hold",
            flags=flags,
            target=target,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "held")


def release(tag, *snapshot, **kwargs):
    """
    Removes a single reference, named with the tag argument, from the
    specified snapshot or snapshots.

    .. note::

        The tag must already exist for each snapshot.
        If a hold exists on a snapshot, attempts to destroy that
        snapshot by using the zfs destroy command return EBUSY.

    tag : string
        name of tag
    snapshot : string
        name of snapshot(s)
    recursive : boolean
        recursively releases a hold with the given tag on the snapshots of
        all descendent file systems.

    .. versionadded:: 2016.3.0
    .. versionchanged:: 2018.3.1

    .. warning::

        As of 2018.3.1 the tag parameter no longer accepts a comma-separated value.
        It's is now possible to create a tag that contains a comma, this was impossible before.

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.release mytag myzpool/mydataset@mysnapshot [recursive=True]
        salt '*' zfs.release mytag myzpool/mydataset@mysnapshot myzpool/mydataset@myothersnapshot

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []
    target = []

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive", False):
        flags.append("-r")

    # NOTE: update target
    target.append(tag)
    target.extend(snapshot)

    ## release snapshot
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="release",
            flags=flags,
            target=target,
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "released")


def snapshot(*snapshot, **kwargs):
    """
    Creates snapshots with the given names.

    snapshot : string
        name of snapshot(s)
    recursive : boolean
        recursively create snapshots of all descendent datasets.
    properties : dict
        additional zfs properties (-o)

    .. note::

        ZFS properties can be specified at the time of creation of the filesystem by
        passing an additional argument called "properties" and specifying the properties
        with their respective values in the form of a python dictionary::

            properties="{'property1': 'value1', 'property2': 'value2'}"

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.snapshot myzpool/mydataset@yesterday [recursive=True]
        salt '*' zfs.snapshot myzpool/mydataset@yesterday myzpool/myotherdataset@yesterday [recursive=True]

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = []

    # NOTE: push filesystem properties
    filesystem_properties = kwargs.get("properties", {})

    # NOTE: set extra config from kwargs
    if kwargs.get("recursive", False):
        flags.append("-r")

    ## Create snapshot
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="snapshot",
            flags=flags,
            filesystem_properties=filesystem_properties,
            target=list(snapshot),
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "snapshotted")


def set(*dataset, **kwargs):
    """
    Sets the property or list of properties to the given value(s) for each dataset.

    dataset : string
        name of snapshot(s), filesystem(s), or volume(s)
    properties : string
        additional zfs properties pairs

    .. note::

        properties are passed as key-value pairs. e.g.

            compression=off

    .. note::

        Only some properties can be edited.

        See the Properties section for more information on what properties
        can be set and acceptable values.

        Numeric values can be specified as exact values, or in a human-readable
        form with a suffix of B, K, M, G, T, P, E (for bytes, kilobytes,
        megabytes, gigabytes, terabytes, petabytes, or exabytes respectively).

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.set myzpool/mydataset compression=off
        salt '*' zfs.set myzpool/mydataset myzpool/myotherdataset compression=off
        salt '*' zfs.set myzpool/mydataset myzpool/myotherdataset compression=lz4 canmount=off

    """
    ## Configure command
    # NOTE: push filesystem properties
    filesystem_properties = salt.utils.args.clean_kwargs(**kwargs)

    ## Set property
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="set",
            property_name=list(filesystem_properties.keys()),
            property_value=list(filesystem_properties.values()),
            target=list(dataset),
        ),
        python_shell=False,
    )

    return __utils__["zfs.parse_command_result"](res, "set")


def get(*dataset, **kwargs):
    """
    Displays properties for the given datasets.

    dataset : string
        name of snapshot(s), filesystem(s), or volume(s)
    properties : string
        comma-separated list of properties to list, defaults to all
    recursive : boolean
        recursively list children
    depth : int
        recursively list children to depth
    fields : string
        comma-separated list of fields to include, the name and property field will always be added
    type : string
        comma-separated list of types to display, where type is one of
        filesystem, snapshot, volume, bookmark, or all.

        .. versionchanged:: 3004

        type is ignored on Solaris 10 and 11 since not a valid parameter on those platforms

    source : string
        comma-separated list of sources to display. Must be one of the following:
        local, default, inherited, temporary, and none. The default value is all sources.
    parsable : boolean
        display numbers in parsable (exact) values (default = True)
        .. versionadded:: 2018.3.0

    .. note::
        If no datasets are specified, then the command displays properties
        for all datasets on the system.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' zfs.get
        salt '*' zfs.get myzpool/mydataset [recursive=True|False]
        salt '*' zfs.get myzpool/mydataset properties="sharenfs,mountpoint" [recursive=True|False]
        salt '*' zfs.get myzpool/mydataset myzpool/myotherdataset properties=available fields=value depth=1

    """
    ## Configure command
    # NOTE: initialize the defaults
    flags = ["-H"]
    opts = {}

    # NOTE: set extra config from kwargs
    if kwargs.get("depth", False):
        opts["-d"] = kwargs.get("depth")
    elif kwargs.get("recursive", False):
        flags.append("-r")
    fields = kwargs.get("fields", "value,source").split(",")
    if "name" in fields:  # ensure name is first
        fields.remove("name")
    if "property" in fields:  # ensure property is second
        fields.remove("property")
    fields.insert(0, "name")
    fields.insert(1, "property")
    opts["-o"] = ",".join(fields)

    if not salt.utils.platform.is_sunos():
        if kwargs.get("type", False):
            opts["-t"] = kwargs.get("type")

    if kwargs.get("source", False):
        opts["-s"] = kwargs.get("source")

    # NOTE: set property_name
    property_name = kwargs.get("properties", "all")

    ## Get properties
    res = __salt__["cmd.run_all"](
        __utils__["zfs.zfs_command"](
            command="get",
            flags=flags,
            opts=opts,
            property_name=property_name,
            target=list(dataset),
        ),
        python_shell=False,
    )

    ret = __utils__["zfs.parse_command_result"](res)
    if res["retcode"] == 0:
        for ds in res["stdout"].splitlines():
            ds_data = OrderedDict(list(zip(fields, ds.split("\t"))))

            if "value" in ds_data:
                if kwargs.get("parsable", True):
                    ds_data["value"] = __utils__["zfs.from_auto"](
                        ds_data["property"],
                        ds_data["value"],
                    )
                else:
                    ds_data["value"] = __utils__["zfs.to_auto"](
                        ds_data["property"],
                        ds_data["value"],
                        convert_to_human=True,
                    )

            if ds_data["name"] not in ret:
                ret[ds_data["name"]] = OrderedDict()

            ret[ds_data["name"]][ds_data["property"]] = ds_data
            del ds_data["name"]
            del ds_data["property"]

    return ret


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
    Unit tests for salt.utils.proxy

    :codeauthor: :email:`Gareth J. Greenaway <gareth@saltstack.com>`
"""


import logging

import salt.grains.esxi as esxi_grains
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class EsxiGrainsTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "__salt__": {},
            "__opts__": {
                "proxy": {
                    "proxytype": "esxi",
                    "host": "esxi.domain.com",
                    "username": "username",
                    "passwords": ["password1"],
                }
            },
            "__pillar__": {
                "proxy": {
                    "proxytype": "esxi",
                    "host": "esxi.domain.com",
                    "username": "username",
                    "passwords": ["password1"],
                }
            },
        }

        return {esxi_grains: module_globals}

    def test_virtual(self):
        with patch("salt.utils.proxy.is_proxytype", return_value=True, autospec=True):
            ret = esxi_grains.__virtual__()
            self.assertEqual(ret, "esxi")

    def test_kernel(self):
        with patch("salt.utils.proxy.is_proxytype", return_value=True, autospec=True):
            ret = esxi_grains.kernel()
            self.assertEqual(ret, {"kernel": "proxy"})

    def test_osfamily(self):
        with patch("salt.utils.proxy.is_proxytype", return_value=True, autospec=True):
            ret = esxi_grains.os_family()
            self.assertEqual(ret, {"os_family": "proxy"})

    def test_os(self):
        grain_cache_return = {
            "name": "VMware vCenter Server",
            "fullName": "VMware vCenter Server 6.7.0 build-15679289",
            "vendor": "VMware, Inc.",
            "version": "6.7.0",
            "build": "15679289",
            "localeVersion": "INTL",
            "localeBuild": "000",
            "osType": "linux-x64",
            "productLineId": "vpx",
            "apiType": "VirtualCenter",
            "apiVersion": "6.7.3",
            "instanceUuid": "058ed113-1820-41dc-a8a9-8a5dd48632a4",
            "licenseProductName": "VMware VirtualCenter Server",
            "licenseProductVersion": "6.0",
        }

        expected = {"os": "VMware vCenter Server 6.7.0 build-15679289"}
        with patch("salt.utils.proxy.is_proxytype", return_value=True, autospec=True):
            with patch(
                "salt.modules.vsphere.system_info", return_value=grain_cache_return
            ):
                ret = esxi_grains.os()
                self.assertEqual(ret, expected)

    def test_esxi(self):
        grain_cache_return = {
            "name": "VMware vCenter Server",
            "fullName": "VMware vCenter Server 6.7.0 build-15679289",
            "vendor": "VMware, Inc.",
            "version": "6.7.0",
            "build": "15679289",
            "localeVersion": "INTL",
            "localeBuild": "000",
            "osType": "linux-x64",
            "productLineId": "vpx",
            "apiType": "VirtualCenter",
            "apiVersion": "6.7.3",
            "instanceUuid": "058ed113-1820-41dc-a8a9-8a5dd48632a4",
            "licenseProductName": "VMware VirtualCenter Server",
            "licenseProductVersion": "6.0",
        }

        with patch("salt.utils.proxy.is_proxytype", return_value=True, autospec=True):
            with patch(
                "salt.modules.vsphere.system_info", return_value=grain_cache_return
            ):
                ret = esxi_grains.esxi()
                self.assertEqual(ret, grain_cache_return)



"""
Test salt.utils.zeromq
"""


import salt.utils.zeromq
import zmq
from salt._compat import ipaddress
from salt.exceptions import SaltSystemExit
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf


class UtilsTestCase(TestCase):
    def test_ip_bracket(self):
        test_ipv4 = "127.0.0.1"
        test_ipv6 = "::1"
        test_ipv6_uri = "[::1]"
        self.assertEqual(test_ipv4, salt.utils.zeromq.ip_bracket(test_ipv4))
        self.assertEqual(
            "[{}]".format(test_ipv6), salt.utils.zeromq.ip_bracket(test_ipv6)
        )
        self.assertEqual(
            "[{}]".format(test_ipv6), salt.utils.zeromq.ip_bracket(test_ipv6_uri)
        )

        ip_addr_obj = ipaddress.ip_address(test_ipv4)
        self.assertEqual(test_ipv4, salt.utils.zeromq.ip_bracket(ip_addr_obj))

    @skipIf(
        not hasattr(zmq, "IPC_PATH_MAX_LEN"), "ZMQ does not have max length support."
    )
    def test_check_ipc_length(self):
        """
        Ensure we throw an exception if we have a too-long IPC URI
        """
        with patch("zmq.IPC_PATH_MAX_LEN", 1):
            self.assertRaises(
                SaltSystemExit, salt.utils.zeromq.check_ipc_path_max_len, "1" * 1024
            )



"""Miscellaneous utility functions and classes.

This module is used internally by Tornado.  It is not necessarily expected
that the functions and classes defined here will be useful to other
applications, but they are documented here in case they are.

The one public-facing part of this module is the `Configurable` class
and its `~Configurable.configure` method, which becomes a part of the
interface of its subclasses, including `.AsyncHTTPClient`, `.IOLoop`,
and `.Resolver`.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import array
import atexit
import os
import re
import sys
import zlib

PY3 = sys.version_info >= (3,)

if PY3:
    xrange = range

# inspect.getargspec() raises DeprecationWarnings in Python 3.5.
# The two functions have compatible interfaces for the parts we need.
if PY3:
    from inspect import getfullargspec as getargspec
else:
    from inspect import getargspec

# Aliases for types that are spelled differently in different Python
# versions. bytes_type is deprecated and no longer used in Tornado
# itself but is left in case anyone outside Tornado is using it.
bytes_type = bytes
if PY3:
    unicode_type = str
    basestring_type = str
else:
    # The names unicode and basestring don't exist in py3 so silence flake8.
    unicode_type = unicode  # noqa
    basestring_type = basestring  # noqa


try:
    import typing  # noqa
    from typing import cast

    _ObjectDictBase = typing.Dict[str, typing.Any]
except ImportError:
    _ObjectDictBase = dict

    def cast(typ, x):
        return x
else:
    # More imports that are only needed in type comments.
    import datetime  # noqa
    import types  # noqa
    from typing import Any, AnyStr, Union, Optional, Dict, Mapping  # noqa
    from typing import Tuple, Match, Callable  # noqa

    if PY3:
        _BaseString = str
    else:
        _BaseString = Union[bytes, unicode_type]


try:
    from sys import is_finalizing
except ImportError:
    # Emulate it
    def _get_emulated_is_finalizing():
        L = []
        atexit.register(lambda: L.append(None))

        def is_finalizing():
            # Not referencing any globals here
            return L != []

        return is_finalizing

    is_finalizing = _get_emulated_is_finalizing()


class ObjectDict(_ObjectDictBase):
    """Makes a dictionary behave like an object, with attribute-style access.
    """
    def __getattr__(self, name):
        # type: (str) -> Any
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        # type: (str, Any) -> None
        self[name] = value


class GzipDecompressor(object):
    """Streaming gzip decompressor.

    The interface is like that of `zlib.decompressobj` (without some of the
    optional arguments, but it understands gzip headers and checksums.
    """
    def __init__(self):
        # Magic parameter makes zlib module understand gzip header
        # http://stackoverflow.com/questions/1838699/how-can-i-decompress-a-gzip-stream-with-zlib
        # This works on cpython and pypy, but not jython.
        self.decompressobj = zlib.decompressobj(16 + zlib.MAX_WBITS)

    def decompress(self, value, max_length=None):
        # type: (bytes, Optional[int]) -> bytes
        """Decompress a chunk, returning newly-available data.

        Some data may be buffered for later processing; `flush` must
        be called when there is no more input data to ensure that
        all data was processed.

        If ``max_length`` is given, some input data may be left over
        in ``unconsumed_tail``; you must retrieve this value and pass
        it back to a future call to `decompress` if it is not empty.
        """
        return self.decompressobj.decompress(value, max_length)

    @property
    def unconsumed_tail(self):
        # type: () -> bytes
        """Returns the unconsumed portion left over
        """
        return self.decompressobj.unconsumed_tail

    def flush(self):
        # type: () -> bytes
        """Return any remaining buffered data not yet returned by decompress.

        Also checks for errors such as truncated input.
        No other methods may be called on this object after `flush`.
        """
        return self.decompressobj.flush()


def import_object(name):
    # type: (_BaseString) -> Any
    """Imports an object by name.

    import_object('x') is equivalent to 'import x'.
    import_object('x.y.z') is equivalent to 'from x.y import z'.

    >>> import tornado.escape
    >>> import_object('tornado.escape') is tornado.escape
    True
    >>> import_object('tornado.escape.utf8') is tornado.escape.utf8
    True
    >>> import_object('tornado') is tornado
    True
    >>> import_object('tornado.missing_module')
    Traceback (most recent call last):
        ...
    ImportError: No module named missing_module
    """
    if not isinstance(name, str):
        # on python 2 a byte string is required.
        name = name.encode('utf-8')
    if name.count('.') == 0:
        return __import__(name, None, None)

    parts = name.split('.')
    obj = __import__('.'.join(parts[:-1]), None, None, [parts[-1]], 0)
    try:
        return getattr(obj, parts[-1])
    except AttributeError:
        raise ImportError("No module named %s" % parts[-1])


# Stubs to make mypy happy (and later for actual type-checking).
def raise_exc_info(exc_info):
    # type: (Tuple[type, BaseException, types.TracebackType]) -> None
    pass


def exec_in(code, glob, loc=None):
    # type: (Any, Dict[str, Any], Optional[Mapping[str, Any]]) -> Any
    if isinstance(code, basestring_type):
        # exec(string) inherits the caller's future imports; compile
        # the string first to prevent that.
        code = compile(code, '<string>', 'exec', dont_inherit=True)
    exec(code, glob, loc)


if PY3:
    exec("""
def raise_exc_info(exc_info):
    try:
        raise exc_info[1].with_traceback(exc_info[2])
    finally:
        exc_info = None

""")
else:
    exec("""
def raise_exc_info(exc_info):
    raise exc_info[0], exc_info[1], exc_info[2]
""")


def errno_from_exception(e):
    # type: (BaseException) -> Optional[int]
    """Provides the errno from an Exception object.

    There are cases that the errno attribute was not set so we pull
    the errno out of the args but if someone instantiates an Exception
    without any args you will get a tuple error. So this function
    abstracts all that behavior to give you a safe way to get the
    errno.
    """

    if hasattr(e, 'errno'):
        return e.errno  # type: ignore
    elif e.args:
        return e.args[0]
    else:
        return None


_alphanum = frozenset(
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789")


def _re_unescape_replacement(match):
    # type: (Match[str]) -> str
    group = match.group(1)
    if group[0] in _alphanum:
        raise ValueError("cannot unescape '\\\\%s'" % group[0])
    return group


_re_unescape_pattern = re.compile(r'\\(.)', re.DOTALL)


def re_unescape(s):
    # type: (str) -> str
    """Unescape a string escaped by `re.escape`.

    May raise ``ValueError`` for regular expressions which could not
    have been produced by `re.escape` (for example, strings containing
    ``\d`` cannot be unescaped).

    .. versionadded:: 4.4
    """
    return _re_unescape_pattern.sub(_re_unescape_replacement, s)


class Configurable(object):
    """Base class for configurable interfaces.

    A configurable interface is an (abstract) class whose constructor
    acts as a factory function for one of its implementation subclasses.
    The implementation subclass as well as optional keyword arguments to
    its initializer can be set globally at runtime with `configure`.

    By using the constructor as the factory method, the interface
    looks like a normal class, `isinstance` works as usual, etc.  This
    pattern is most useful when the choice of implementation is likely
    to be a global decision (e.g. when `~select.epoll` is available,
    always use it instead of `~select.select`), or when a
    previously-monolithic class has been split into specialized
    subclasses.

    Configurable subclasses must define the class methods
    `configurable_base` and `configurable_default`, and use the instance
    method `initialize` instead of ``__init__``.
    """
    __impl_class = None  # type: type
    __impl_kwargs = None  # type: Dict[str, Any]

    def __new__(cls, *args, **kwargs):
        base = cls.configurable_base()
        init_kwargs = {}
        if cls is base:
            impl = cls.configured_class()
            if base.__impl_kwargs:
                init_kwargs.update(base.__impl_kwargs)
        else:
            impl = cls
        init_kwargs.update(kwargs)
        instance = super(Configurable, cls).__new__(impl)
        # initialize vs __init__ chosen for compatibility with AsyncHTTPClient
        # singleton magic.  If we get rid of that we can switch to __init__
        # here too.
        instance.initialize(*args, **init_kwargs)
        return instance

    @classmethod
    def configurable_base(cls):
        # type: () -> Any
        # TODO: This class needs https://github.com/python/typing/issues/107
        # to be fully typeable.
        """Returns the base class of a configurable hierarchy.

        This will normally return the class in which it is defined.
        (which is *not* necessarily the same as the cls classmethod parameter).
        """
        raise NotImplementedError()

    @classmethod
    def configurable_default(cls):
        # type: () -> type
        """Returns the implementation class to be used if none is configured."""
        raise NotImplementedError()

    def initialize(self):
        # type: () -> None
        """Initialize a `Configurable` subclass instance.

        Configurable classes should use `initialize` instead of ``__init__``.

        .. versionchanged:: 4.2
           Now accepts positional arguments in addition to keyword arguments.
        """

    @classmethod
    def configure(cls, impl, **kwargs):
        # type: (Any, **Any) -> None
        """Sets the class to use when the base class is instantiated.

        Keyword arguments will be saved and added to the arguments passed
        to the constructor.  This can be used to set global defaults for
        some parameters.
        """
        base = cls.configurable_base()
        if isinstance(impl, (str, unicode_type)):
            impl = import_object(impl)
        if impl is not None and not issubclass(impl, cls):
            raise ValueError("Invalid subclass of %s" % cls)
        base.__impl_class = impl
        base.__impl_kwargs = kwargs

    @classmethod
    def configured_class(cls):
        # type: () -> type
        """Returns the currently configured class."""
        base = cls.configurable_base()
        if cls.__impl_class is None:
            base.__impl_class = cls.configurable_default()
        return base.__impl_class

    @classmethod
    def _save_configuration(cls):
        # type: () -> Tuple[type, Dict[str, Any]]
        base = cls.configurable_base()
        return (base.__impl_class, base.__impl_kwargs)

    @classmethod
    def _restore_configuration(cls, saved):
        # type: (Tuple[type, Dict[str, Any]]) -> None
        base = cls.configurable_base()
        base.__impl_class = saved[0]
        base.__impl_kwargs = saved[1]


class ArgReplacer(object):
    """Replaces one value in an ``args, kwargs`` pair.

    Inspects the function signature to find an argument by name
    whether it is passed by position or keyword.  For use in decorators
    and similar wrappers.
    """
    def __init__(self, func, name):
        # type: (Callable, str) -> None
        self.name = name
        try:
            self.arg_pos = self._getargnames(func).index(name)
        except ValueError:
            # Not a positional parameter
            self.arg_pos = None

    def _getargnames(self, func):
        # type: (Callable) -> List[str]
        try:
            return getargspec(func).args
        except TypeError:
            if hasattr(func, 'func_code'):
                # Cython-generated code has all the attributes needed
                # by inspect.getargspec, but the inspect module only
                # works with ordinary functions. Inline the portion of
                # getargspec that we need here. Note that for static
                # functions the @cython.binding(True) decorator must
                # be used (for methods it works out of the box).
                code = func.func_code  # type: ignore
                return code.co_varnames[:code.co_argcount]
            raise

    def get_old_value(self, args, kwargs, default=None):
        # type: (List[Any], Dict[str, Any], Any) -> Any
        """Returns the old value of the named argument without replacing it.

        Returns ``default`` if the argument is not present.
        """
        if self.arg_pos is not None and len(args) > self.arg_pos:
            return args[self.arg_pos]
        else:
            return kwargs.get(self.name, default)

    def replace(self, new_value, args, kwargs):
        # type: (Any, List[Any], Dict[str, Any]) -> Tuple[Any, List[Any], Dict[str, Any]]
        """Replace the named argument in ``args, kwargs`` with ``new_value``.

        Returns ``(old_value, args, kwargs)``.  The returned ``args`` and
        ``kwargs`` objects may not be the same as the input objects, or
        the input objects may be mutated.

        If the named argument was not found, ``new_value`` will be added
        to ``kwargs`` and None will be returned as ``old_value``.
        """
        if self.arg_pos is not None and len(args) > self.arg_pos:
            # The arg to replace is passed positionally
            old_value = args[self.arg_pos]
            args = list(args)  # *args is normally a tuple
            args[self.arg_pos] = new_value
        else:
            # The arg to replace is either omitted or passed by keyword.
            old_value = kwargs.get(self.name)
            kwargs[self.name] = new_value
        return old_value, args, kwargs


def timedelta_to_seconds(td):
    # type: (datetime.timedelta) -> float
    """Equivalent to td.total_seconds() (introduced in python 2.7)."""
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10 ** 6) / float(10 ** 6)


def _websocket_mask_python(mask, data):
    # type: (bytes, bytes) -> bytes
    """Websocket masking function.

    `mask` is a `bytes` object of length 4; `data` is a `bytes` object of any length.
    Returns a `bytes` object of the same length as `data` with the mask applied
    as specified in section 5.3 of RFC 6455.

    This pure-python implementation may be replaced by an optimized version when available.
    """
    mask_arr = array.array("B", mask)
    unmasked_arr = array.array("B", data)
    for i in xrange(len(data)):
        unmasked_arr[i] = unmasked_arr[i] ^ mask_arr[i % 4]
    if PY3:
        # tostring was deprecated in py32.  It hasn't been removed,
        # but since we turn on deprecation warnings in our tests
        # we need to use the right one.
        return unmasked_arr.tobytes()
    else:
        return unmasked_arr.tostring()


if (os.environ.get('TORNADO_NO_EXTENSION') or
        os.environ.get('TORNADO_EXTENSION') == '0'):
    # These environment variables exist to make it easier to do performance
    # comparisons; they are not guaranteed to remain supported in the future.
    _websocket_mask = _websocket_mask_python
else:
    try:
        from salt.ext.tornado.speedups import websocket_mask as _websocket_mask
    except ImportError:
        if os.environ.get('TORNADO_EXTENSION') == '1':
            raise
        _websocket_mask = _websocket_mask_python


def doctests():
    import doctest
    return doctest.DocTestSuite()



import salt.modules.selinux as selinux
from salt.exceptions import SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SelinuxModuleTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.selinux
    """

    def setup_loader_modules(self):
        return {selinux: {}}

    def test_fcontext_get_policy_parsing(self):
        """
        Test to verify that the parsing of the semanage output into fields is
        correct. Added with #45784.
        """
        cases = [
            {
                "semanage_out": (
                    "/var/www(/.*)?     all files         "
                    " system_u:object_r:httpd_sys_content_t:s0"
                ),
                "name": "/var/www(/.*)?",
                "filetype": "all files",
                "sel_user": "system_u",
                "sel_role": "object_r",
                "sel_type": "httpd_sys_content_t",
                "sel_level": "s0",
            },
            {
                "semanage_out": (
                    "/var/www(/.*)? all files         "
                    " system_u:object_r:httpd_sys_content_t:s0"
                ),
                "name": "/var/www(/.*)?",
                "filetype": "all files",
                "sel_user": "system_u",
                "sel_role": "object_r",
                "sel_type": "httpd_sys_content_t",
                "sel_level": "s0",
            },
            {
                "semanage_out": (
                    "/var/lib/dhcp3?                                    directory      "
                    "    system_u:object_r:dhcp_state_t:s0"
                ),
                "name": "/var/lib/dhcp3?",
                "filetype": "directory",
                "sel_user": "system_u",
                "sel_role": "object_r",
                "sel_type": "dhcp_state_t",
                "sel_level": "s0",
            },
            {
                "semanage_out": (
                    "/var/lib/dhcp3?  directory         "
                    " system_u:object_r:dhcp_state_t:s0"
                ),
                "name": "/var/lib/dhcp3?",
                "filetype": "directory",
                "sel_user": "system_u",
                "sel_role": "object_r",
                "sel_type": "dhcp_state_t",
                "sel_level": "s0",
            },
            {
                "semanage_out": (
                    "/var/lib/dhcp3? directory         "
                    " system_u:object_r:dhcp_state_t:s0"
                ),
                "name": "/var/lib/dhcp3?",
                "filetype": "directory",
                "sel_user": "system_u",
                "sel_role": "object_r",
                "sel_type": "dhcp_state_t",
                "sel_level": "s0",
            },
        ]

        for case in cases:
            with patch.dict(
                selinux.__salt__,
                {"cmd.shell": MagicMock(return_value=case["semanage_out"])},
            ):
                ret = selinux.fcontext_get_policy(case["name"])
                self.assertEqual(ret["filespec"], case["name"])
                self.assertEqual(ret["filetype"], case["filetype"])
                self.assertEqual(ret["sel_user"], case["sel_user"])
                self.assertEqual(ret["sel_role"], case["sel_role"])
                self.assertEqual(ret["sel_type"], case["sel_type"])
                self.assertEqual(ret["sel_level"], case["sel_level"])

    def test_parse_protocol_port_positive(self):
        """
        Test to verify positive parsing name, protocol and port combinations
        """
        cases = [
            {
                "name": "tcp/80",
                "protocol": None,
                "port": None,
                "expected": ("tcp", "80"),
            },
            {
                "name": "udp/53",
                "protocol": None,
                "port": None,
                "expected": ("udp", "53"),
            },
            {
                "name": "tcp_test_dns",
                "protocol": "tcp",
                "port": "53",
                "expected": ("tcp", "53"),
            },
            {
                "name": "udp_test/dns",
                "protocol": "udp",
                "port": "53",
                "expected": ("udp", "53"),
            },
        ]

        for case in cases:
            ret = selinux._parse_protocol_port(
                case["name"], case["protocol"], case["port"]
            )
            self.assertTupleEqual(ret, case["expected"])

    def test_parse_protocol_port_negative(self):
        """
        Test to verify negative parsing of name, protocol and port combinations
        """
        cases = [
            {"name": "invalid_name_no_args", "protocol": None, "port": None},
            {"name": "invalid_proto/80", "protocol": "nottcp", "port": "80"},
            {"name": "invalid_port", "protocol": "tcp", "port": "notaport"},
            {"name": "missing_proto", "protocol": None, "port": "80"},
            {"name": "missing_port", "protocol": "udp", "port": None},
        ]

        for case in cases:
            self.assertRaises(
                SaltInvocationError,
                selinux._parse_protocol_port,
                case["name"],
                case["protocol"],
                case["port"],
            )

    def test_port_get_policy_parsing(self):
        """
        Test to verify that the parsing of the semanage port output into fields is correct.
        """
        cases = [
            {
                "semanage_out": "cma_port_t                     tcp      1050",
                "name": "tcp/1050",
                "expected": {
                    "sel_type": "cma_port_t",
                    "protocol": "tcp",
                    "port": "1050",
                },
            },
            {
                "semanage_out": (
                    "cluster_port_t                 tcp      5149, 40040, 50006-50008"
                ),
                "name": "tcp/40040",
                "expected": {
                    "sel_type": "cluster_port_t",
                    "protocol": "tcp",
                    "port": "5149, 40040, 50006-50008",
                },
            },
            {
                "semanage_out": (
                    "http_port_t                    tcp      9008, 8010, 9002-9003, 80,"
                    " 81, 443, 488, 8008, 8009, 8443, 9000"
                ),
                "name": "tcp/9000",
                "expected": {
                    "sel_type": "http_port_t",
                    "protocol": "tcp",
                    "port": (
                        "9008, 8010, 9002-9003, 80, 81, 443, 488, 8008, 8009, 8443,"
                        " 9000"
                    ),
                },
            },
            {
                "semanage_out": (
                    "vnc_port_t                     tcp      5985-5999, 5900-5983"
                ),
                "name": "tcp/5985-5999",
                "expected": {
                    "sel_type": "vnc_port_t",
                    "protocol": "tcp",
                    "port": "5985-5999, 5900-5983",
                },
            },
            {
                "semanage_out": (
                    "zebra_port_t                   tcp      2606, 2608-2609, 2600-2604"
                ),
                "name": "tcp/2608-2609",
                "expected": {
                    "sel_type": "zebra_port_t",
                    "protocol": "tcp",
                    "port": "2606, 2608-2609, 2600-2604",
                },
            },
            {
                "semanage_out": (
                    "radius_port_t                  udp      1645, 1812, 18120-18121"
                ),
                "name": "tcp/18120-18121",
                "expected": {
                    "sel_type": "radius_port_t",
                    "protocol": "udp",
                    "port": "1645, 1812, 18120-18121",
                },
            },
        ]

        for case in cases:
            with patch.dict(
                selinux.__salt__,
                {"cmd.shell": MagicMock(return_value=case["semanage_out"])},
            ):
                ret = selinux.port_get_policy(case["name"])
                self.assertDictEqual(ret, case["expected"])

    def test_fcontext_policy_parsing_new(self):
        """
        Test parsing the stdout response of restorecon used in fcontext_policy_applied, new style.
        """
        restorecon_ret = (
            "Would relabel /foo/bar from some_u:some_r:some_t:s0 to"
            " other_u:other_r:other_t:s0"
        )
        with patch.object(
            selinux, "fcontext_policy_is_applied", return_value=restorecon_ret
        ), patch.dict(
            selinux.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertEqual(
                selinux.fcontext_apply_policy("/foo/bar"),
                {
                    "changes": {
                        "/foo/bar": {
                            "old": {
                                "sel_role": "some_r",
                                "sel_type": "some_t",
                                "sel_user": "some_u",
                            },
                            "new": {
                                "sel_role": "other_r",
                                "sel_type": "other_t",
                                "sel_user": "other_u",
                            },
                        },
                    },
                    "retcode": 0,
                },
            )

    def test_fcontext_policy_parsing_old(self):
        """
        Test parsing the stdout response of restorecon used in fcontext_policy_applied, old style.
        """
        restorecon_ret = (
            "restorecon reset /foo/bar context"
            " some_u:some_r:some_t:s0->other_u:other_r:other_t:s0"
        )
        with patch.object(
            selinux, "fcontext_policy_is_applied", return_value=restorecon_ret
        ), patch.dict(
            selinux.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertEqual(
                selinux.fcontext_apply_policy("/foo/bar"),
                {
                    "changes": {
                        "/foo/bar": {
                            "old": {
                                "sel_role": "some_r",
                                "sel_type": "some_t",
                                "sel_user": "some_u",
                            },
                            "new": {
                                "sel_role": "other_r",
                                "sel_type": "other_t",
                                "sel_user": "other_u",
                            },
                        },
                    },
                    "retcode": 0,
                },
            )

    def test_fcontext_policy_parsing_fail(self):
        """
        Test failure response for invalid restorecon data.
        """
        restorecon_ret = "And now for something completely different."
        with patch.object(
            selinux, "fcontext_policy_is_applied", return_value=restorecon_ret
        ), patch.dict(
            selinux.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            self.assertEqual(
                selinux.fcontext_apply_policy("/foo/bar"),
                {
                    "retcode": 1,
                    "error": "Unrecognized response from restorecon command.",
                },
            )



"""
Module to provide RabbitMQ compatibility to Salt.
Todo: A lot, need to add cluster support, logging, and minion configuration
data.
"""

import logging
import os
import random
import re
import string

import salt.utils.itertools
import salt.utils.json
import salt.utils.path
import salt.utils.platform
import salt.utils.user
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.utils.versions import LooseVersion as _LooseVersion

log = logging.getLogger(__name__)

RABBITMQCTL = None
RABBITMQ_PLUGINS = None


def __virtual__():
    """
    Verify RabbitMQ is installed.
    """
    global RABBITMQCTL
    global RABBITMQ_PLUGINS

    if salt.utils.platform.is_windows():
        import winreg

        key = None
        try:
            key = winreg.OpenKeyEx(
                winreg.HKEY_LOCAL_MACHINE,
                "SOFTWARE\\VMware, Inc.\\RabbitMQ Server",
                0,
                winreg.KEY_READ | winreg.KEY_WOW64_32KEY,
            )
            (dir_path, value_type) = winreg.QueryValueEx(key, "Install_Dir")
            if value_type != winreg.REG_SZ:
                raise TypeError(
                    "Invalid RabbitMQ Server directory type: {}".format(value_type)
                )
            if not os.path.isdir(dir_path):
                raise OSError("RabbitMQ directory not found: {}".format(dir_path))
            subdir_match = ""
            for name in os.listdir(dir_path):
                if name.startswith("rabbitmq_server-"):
                    subdir_path = os.path.join(dir_path, name)
                    # Get the matching entry that is last in ASCII order.
                    if os.path.isdir(subdir_path) and subdir_path > subdir_match:
                        subdir_match = subdir_path
            if not subdir_match:
                raise OSError(
                    '"rabbitmq_server-*" subdirectory not found in: {}'.format(dir_path)
                )
            RABBITMQCTL = os.path.join(subdir_match, "sbin", "rabbitmqctl.bat")
            RABBITMQ_PLUGINS = os.path.join(
                subdir_match, "sbin", "rabbitmq-plugins.bat"
            )
        except Exception:  # pylint: disable=broad-except
            pass
        finally:
            if key is not None:
                winreg.CloseKey(key)
    else:
        RABBITMQCTL = salt.utils.path.which("rabbitmqctl")
        RABBITMQ_PLUGINS = salt.utils.path.which("rabbitmq-plugins")

    if not RABBITMQCTL:
        return (False, "Module rabbitmq: module only works when RabbitMQ is installed")
    return True


def _check_response(response):
    if isinstance(response, dict):
        if response["retcode"] != 0 or response["stderr"]:
            raise CommandExecutionError(
                "RabbitMQ command failed: {}".format(response["stderr"])
            )
    else:
        if "Error" in response:
            raise CommandExecutionError("RabbitMQ command failed: {}".format(response))


def _format_response(response, msg):
    if isinstance(response, dict):
        if response["retcode"] != 0 or response["stderr"]:
            raise CommandExecutionError(
                "RabbitMQ command failed: {}".format(response["stderr"])
            )
        else:
            response = response["stdout"]
    else:
        if "Error" in response:
            raise CommandExecutionError("RabbitMQ command failed: {}".format(response))
    return {msg: response}


def _get_rabbitmq_plugin():
    """
    Returns the rabbitmq-plugin command path if we're running an OS that
    doesn't put it in the standard /usr/bin or /usr/local/bin
    This works by taking the rabbitmq-server version and looking for where it
    seems to be hidden in /usr/lib.
    """
    global RABBITMQ_PLUGINS

    if RABBITMQ_PLUGINS is None:
        version = __salt__["pkg.version"]("rabbitmq-server").split("-")[0]
        RABBITMQ_PLUGINS = (
            "/usr/lib/rabbitmq/lib/rabbitmq_server-{}/sbin/rabbitmq-plugins".format(
                version
            )
        )

    return RABBITMQ_PLUGINS


def _safe_output(line):
    """
    Looks for rabbitmqctl warning, or general formatting, strings that aren't
    intended to be parsed as output.
    Returns a boolean whether the line can be parsed as rabbitmqctl output.
    """
    return not any(
        [
            line.startswith("Listing") and line.endswith("..."),
            line.startswith("Listing") and "\t" not in line,
            "...done" in line,
            line.startswith("WARNING:"),
            len(line) == 0,
        ]
    )


def _strip_listing_to_done(output_list):
    """
    Conditionally remove non-relevant first and last line,
    "Listing ..." - "...done".
    outputlist: rabbitmq command output split by newline
    return value: list, conditionally modified, may be empty.
    """
    return [line for line in output_list if _safe_output(line)]


def _output_to_dict(cmdoutput, values_mapper=None):
    """
    Convert rabbitmqctl output to a dict of data
    cmdoutput: string output of rabbitmqctl commands
    values_mapper: function object to process the values part of each line
    """
    if isinstance(cmdoutput, dict):
        if cmdoutput["retcode"] != 0 or cmdoutput["stderr"]:
            raise CommandExecutionError(
                "RabbitMQ command failed: {}".format(cmdoutput["stderr"])
            )
        cmdoutput = cmdoutput["stdout"]

    ret = {}
    if values_mapper is None:
        values_mapper = lambda string: string.split("\t")

    # remove first and last line: Listing ... - ...done
    data_rows = _strip_listing_to_done(cmdoutput.splitlines())

    for row in data_rows:
        try:
            key, values = row.split("\t", 1)
        except ValueError:
            # If we have reached this far, we've hit an edge case where the row
            # only has one item: the key. The key doesn't have any values, so we
            # set it to an empty string to preserve rabbitmq reporting behavior.
            # e.g. A user's permission string for '/' is set to ['', '', ''],
            # Rabbitmq reports this only as '/' from the rabbitmqctl command.
            log.debug(
                "Could not find any values for key '%s'. "
                "Setting to '%s' to an empty string.",
                row,
                row,
            )
            ret[row] = ""
            continue
        ret[key] = values_mapper(values)
    return ret


def _output_to_list(cmdoutput):
    """
    Convert rabbitmqctl output to a list of strings (assuming whitespace-delimited output).
    Ignores output lines that shouldn't be parsed, like warnings.
    cmdoutput: string output of rabbitmqctl commands
    """
    return [
        item
        for line in cmdoutput.splitlines()
        if _safe_output(line)
        for item in line.split()
    ]


def _output_lines_to_list(cmdoutput):
    """
    Convert rabbitmqctl output to a list of strings (assuming newline-delimited output).
    Ignores output lines that shouldn't be parsed, like warnings.
    cmdoutput: string output of rabbitmqctl commands
    """
    return [line.strip() for line in cmdoutput.splitlines() if _safe_output(line)]


def list_users(runas=None):
    """
    Return a list of users based off of rabbitmqctl user_list.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_users
    """
    # Windows runas currently requires a password.
    # Due to this, don't use a default value for
    # runas in Windows.
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "list_users", "-q"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )

    # func to get tags from string such as "[admin, monitoring]"
    func = (
        lambda string: [x.strip() for x in string[1:-1].split(",")]
        if "," in string
        else [x for x in string[1:-1].split(" ")]
    )
    return _output_to_dict(res, func)


def list_vhosts(runas=None):
    """
    Return a list of vhost based on rabbitmqctl list_vhosts.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_vhosts
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "list_vhosts", "-q"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return _output_to_list(res["stdout"])


def list_upstreams(runas=None):
    """
    Returns a dict of upstreams based on rabbitmqctl list_parameters.

    :param str runas: The name of the user to run this command as.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_upstreams

    .. versionadded:: 3000
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    ret = {}
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "list_parameters", "-q"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    for raw_line in res["stdout"].split("\n"):
        if _safe_output(raw_line):
            (_, name, definition) = raw_line.split("\t")
            ret[name] = definition
    return ret


def user_exists(name, runas=None):
    """
    Return whether the user exists based on rabbitmqctl list_users.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.user_exists rabbit_user
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    return name in list_users(runas=runas)


def vhost_exists(name, runas=None):
    """
    Return whether the vhost exists based on rabbitmqctl list_vhosts.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.vhost_exists rabbit_host
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    return name in list_vhosts(runas=runas)


def upstream_exists(name, runas=None):
    """
    Return whether the upstreamexists based on rabbitmqctl list_parameters.

    :param str name: The name of the upstream to check for.
    :param str runas: The name of the user to run the command as.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.upstream_exists rabbit_upstream

    .. versionadded:: 3000
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    return name in list_upstreams(runas=runas)


def add_user(name, password=None, runas=None):
    """
    Add a rabbitMQ user via rabbitmqctl user_add <user> <password>

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.add_user rabbit_user password
    """
    clear_pw = False

    if password is None:
        # Generate a random, temporary password. RabbitMQ requires one.
        clear_pw = True
        password = "".join(
            random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for x in range(15)
        )

    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()

    if salt.utils.platform.is_windows():
        # On Windows, if the password contains a special character
        # such as '|', normal execution will fail. For example:
        # cmd: rabbitmq.add_user abc "asdf|def"
        # stderr: 'def' is not recognized as an internal or external
        #         command,\r\noperable program or batch file.
        # Work around this by using a shell and a quoted command.
        python_shell = True
        cmd = '"{}" add_user "{}" "{}"'.format(RABBITMQCTL, name, password)
    else:
        python_shell = False
        cmd = [RABBITMQCTL, "add_user", name, password]

    res = __salt__["cmd.run_all"](
        cmd,
        reset_system_locale=False,
        output_loglevel="quiet",
        runas=runas,
        python_shell=python_shell,
    )

    if clear_pw:
        # Now, Clear the random password from the account, if necessary
        try:
            clear_password(name, runas)
        except Exception:  # pylint: disable=broad-except
            # Clearing the password failed. We should try to cleanup
            # and rerun and error.
            delete_user(name, runas)
            raise

    msg = "Added"
    return _format_response(res, msg)


def delete_user(name, runas=None):
    """
    Deletes a user via rabbitmqctl delete_user.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.delete_user rabbit_user
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "delete_user", name],
        reset_system_locale=False,
        python_shell=False,
        runas=runas,
    )
    msg = "Deleted"

    return _format_response(res, msg)


def change_password(name, password, runas=None):
    """
    Changes a user's password.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.change_password rabbit_user password
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    if salt.utils.platform.is_windows():
        # On Windows, if the password contains a special character
        # such as '|', normal execution will fail. For example:
        # cmd: rabbitmq.add_user abc "asdf|def"
        # stderr: 'def' is not recognized as an internal or external
        #         command,\r\noperable program or batch file.
        # Work around this by using a shell and a quoted command.
        python_shell = True
        cmd = '"{}" change_password "{}" "{}"'.format(RABBITMQCTL, name, password)
    else:
        python_shell = False
        cmd = [RABBITMQCTL, "change_password", name, password]
    res = __salt__["cmd.run_all"](
        cmd,
        reset_system_locale=False,
        runas=runas,
        output_loglevel="quiet",
        python_shell=python_shell,
    )
    msg = "Password Changed"

    return _format_response(res, msg)


def clear_password(name, runas=None):
    """
    Removes a user's password.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.clear_password rabbit_user
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "clear_password", name],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    msg = "Password Cleared"

    return _format_response(res, msg)


def check_password(name, password, runas=None):
    """
    .. versionadded:: 2016.3.0

    Checks if a user's password is valid.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.check_password rabbit_user password
    """
    # try to get the rabbitmq-version - adapted from _get_rabbitmq_plugin

    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()

    try:
        res = __salt__["cmd.run"](
            [RABBITMQCTL, "status"],
            reset_system_locale=False,
            runas=runas,
            python_shell=False,
        )
        # Check regex against older RabbitMQ version status output
        old_server_version = re.search(r'\{rabbit,"RabbitMQ","(.+)"\}', res)
        # Check regex against newer RabbitMQ version status output
        server_version = re.search(r"RabbitMQ version:\s*(.+)", res)

        if server_version is None and old_server_version is None:
            raise ValueError

        if old_server_version:
            server_version = old_server_version
        server_version = server_version.group(1).split("-")[0]
        version = [int(i) for i in server_version.split(".")]

    except ValueError:
        version = (0, 0, 0)
    if len(version) < 3:
        version = (0, 0, 0)

    # rabbitmq introduced a native api to check a username and password in version 3.5.7.
    if tuple(version) >= (3, 5, 7):
        if salt.utils.platform.is_windows():
            # On Windows, if the password contains a special character
            # such as '|', normal execution will fail. For example:
            # cmd: rabbitmq.add_user abc "asdf|def"
            # stderr: 'def' is not recognized as an internal or external
            #         command,\r\noperable program or batch file.
            # Work around this by using a shell and a quoted command.
            python_shell = True
            cmd = '"{}" authenticate_user "{}" "{}"'.format(RABBITMQCTL, name, password)
        else:
            python_shell = False
            cmd = [RABBITMQCTL, "authenticate_user", name, password]

        res = __salt__["cmd.run_all"](
            cmd,
            reset_system_locale=False,
            runas=runas,
            output_loglevel="quiet",
            python_shell=python_shell,
        )

        if res["retcode"] != 0 or res["stderr"]:
            return False
        return True

    cmd = (
        "rabbit_auth_backend_internal:check_user_login"
        '(<<"{}">>, [{{password, <<"{}">>}}]).'.format(
            name.replace('"', '\\"'), password.replace('"', '\\"')
        )
    )

    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "eval", cmd],
        reset_system_locale=False,
        runas=runas,
        output_loglevel="quiet",
        python_shell=False,
    )
    msg = "password-check"

    _response = _format_response(res, msg)
    _key = next(iter(_response))

    if "invalid credentials" in _response[_key]:
        return False

    return True


def add_vhost(vhost, runas=None):
    """
    Adds a vhost via rabbitmqctl add_vhost.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq add_vhost '<vhost_name>'
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "add_vhost", vhost],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )

    msg = "Added"
    return _format_response(res, msg)


def delete_vhost(vhost, runas=None):
    """
    Deletes a vhost rabbitmqctl delete_vhost.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.delete_vhost '<vhost_name>'
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "delete_vhost", vhost],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    msg = "Deleted"
    return _format_response(res, msg)


def set_permissions(vhost, user, conf=".*", write=".*", read=".*", runas=None):
    """
    Sets permissions for vhost via rabbitmqctl set_permissions

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.set_permissions myvhost myuser
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "set_permissions", "-p", vhost, user, conf, write, read],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    msg = "Permissions Set"
    return _format_response(res, msg)


def list_permissions(vhost, runas=None):
    """
    Lists permissions for vhost via rabbitmqctl list_permissions

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_permissions /myvhost
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "list_permissions", "--formatter=json", "-p", vhost],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )

    perms = salt.utils.json.loads(res["stdout"])
    perms_dict = {}
    for perm in perms:
        user = perm["user"]
        perms_dict[user] = perm
        del perms_dict[user]["user"]
    return perms_dict


def list_user_permissions(name, runas=None):
    """
    List permissions for a user via rabbitmqctl list_user_permissions

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_user_permissions user
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "list_user_permissions", name, "--formatter=json"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )

    perms = salt.utils.json.loads(res["stdout"])
    perms_dict = {}
    for perm in perms:
        vhost = perm["vhost"]
        perms_dict[vhost] = perm
        del perms_dict[vhost]["vhost"]
    return perms_dict


def set_user_tags(name, tags, runas=None):
    """Add user tags via rabbitmqctl set_user_tags

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.set_user_tags myadmin administrator
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()

    if not isinstance(tags, (list, tuple)):
        tags = [tags]

    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "set_user_tags", name] + list(tags),
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    msg = "Tag(s) set"
    return _format_response(res, msg)


def status(runas=None):
    """
    return rabbitmq status

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.status
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "status"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return res["stdout"]


def cluster_status(runas=None):
    """
    return rabbitmq cluster_status

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.cluster_status
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "cluster_status"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return res["stdout"]


def join_cluster(host, user="rabbit", ram_node=None, runas=None):
    """
    Join a rabbit cluster

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.join_cluster rabbit.example.com rabbit
    """
    cmd = [RABBITMQCTL, "join_cluster"]
    if ram_node:
        cmd.append("--ram")
    cmd.append("{}@{}".format(user, host))

    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    stop_app(runas)
    res = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    start_app(runas)

    return _format_response(res, "Join")


def stop_app(runas=None):
    """
    Stops the RabbitMQ application, leaving the Erlang node running.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.stop_app
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "stop_app"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return res["stdout"]


def start_app(runas=None):
    """
    Start the RabbitMQ application.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.start_app
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "start_app"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return res["stdout"]


def reset(runas=None):
    """
    Return a RabbitMQ node to its virgin state

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.reset
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "reset"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return res["stdout"]


def force_reset(runas=None):
    """
    Forcefully Return a RabbitMQ node to its virgin state

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.force_reset
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "force_reset"],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return res["stdout"]


def list_queues(runas=None, *args):
    """
    Returns queue details of the / virtual host

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_queues messages consumers
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    cmd = [RABBITMQCTL, "list_queues", "-q"]
    cmd.extend(args)
    res = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    _check_response(res)
    return _output_to_dict(res["stdout"])


def list_queues_vhost(vhost, runas=None, *args):
    """
    Returns queue details of specified virtual host. This command will consider
    first parameter as the vhost name and rest will be treated as
    queueinfoitem. For getting details on vhost ``/``, use :mod:`list_queues
    <salt.modules.rabbitmq.list_queues>` instead).

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_queues messages consumers
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    cmd = [RABBITMQCTL, "list_queues", "-q", "-p", vhost]
    cmd.extend(args)
    res = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    _check_response(res)
    return _output_to_dict(res["stdout"])


def list_policies(vhost="/", runas=None):
    """
    Return a dictionary of policies nested by vhost and name
    based on the data returned from rabbitmqctl list_policies.

    Reference: http://www.rabbitmq.com/ha.html

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_policies
    """
    ret = {}
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "list_policies", "-q", "-p", vhost],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    output = res["stdout"]

    if __grains__["os_family"] != "FreeBSD":
        version = __salt__["pkg.version"]("rabbitmq-server").split("-")[0]
    else:
        version = __salt__["pkg.version"]("rabbitmq").split("-")[0]

    for line in _output_lines_to_list(output):
        parts = line.split("\t")

        if len(parts) not in (5, 6):
            continue

        vhost, name = parts[0], parts[1]
        if vhost not in ret:
            ret[vhost] = {}
        ret[vhost][name] = {}

        if _LooseVersion(version) >= _LooseVersion("3.7"):
            # in version 3.7 the position of apply_to and pattern has been
            # switched
            ret[vhost][name]["pattern"] = parts[2]
            ret[vhost][name]["apply_to"] = parts[3]
            ret[vhost][name]["definition"] = parts[4]
            ret[vhost][name]["priority"] = parts[5]
        else:
            # How many fields are there? - 'apply_to' was inserted in position
            # 2 at some point
            # and in version 3.7 the position of apply_to and pattern has been
            # switched
            offset = len(parts) - 5
            if len(parts) == 6:
                ret[vhost][name]["apply_to"] = parts[2]
            ret[vhost][name].update(
                {
                    "pattern": parts[offset + 2],
                    "definition": parts[offset + 3],
                    "priority": parts[offset + 4],
                }
            )

    return ret


def set_policy(
    vhost, name, pattern, definition, priority=None, runas=None, apply_to=None
):
    """
    Set a policy based on rabbitmqctl set_policy.

    Reference: http://www.rabbitmq.com/ha.html

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.set_policy / HA '.*' '{"ha-mode":"all"}'
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    if isinstance(definition, dict):
        definition = salt.utils.json.dumps(definition)
    if not isinstance(definition, str):
        raise SaltInvocationError(
            "The 'definition' argument must be a dictionary or JSON string"
        )
    cmd = [RABBITMQCTL, "set_policy", "-p", vhost]
    if priority:
        cmd.extend(["--priority", priority])
    if apply_to:
        cmd.extend(["--apply-to", apply_to])
    cmd.extend([name, pattern, definition])
    res = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    log.debug("Set policy: %s", res["stdout"])
    return _format_response(res, "Set")


def delete_policy(vhost, name, runas=None):
    """
    Delete a policy based on rabbitmqctl clear_policy.

    Reference: http://www.rabbitmq.com/ha.html

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.delete_policy / HA
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "clear_policy", "-p", vhost, name],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    log.debug("Delete policy: %s", res["stdout"])
    return _format_response(res, "Deleted")


def policy_exists(vhost, name, runas=None):
    """
    Return whether the policy exists based on rabbitmqctl list_policies.

    Reference: http://www.rabbitmq.com/ha.html

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.policy_exists / HA
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    policies = list_policies(runas=runas)
    return bool(vhost in policies and name in policies[vhost])


def list_available_plugins(runas=None):
    """
    Returns a list of the names of all available plugins (enabled and disabled).

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_available_plugins
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    cmd = [_get_rabbitmq_plugin(), "list", "-m"]
    ret = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    _check_response(ret)
    return _output_to_list(ret["stdout"])


def list_enabled_plugins(runas=None):
    """
    Returns a list of the names of the enabled plugins.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.list_enabled_plugins
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    cmd = [_get_rabbitmq_plugin(), "list", "-m", "-e"]
    ret = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    _check_response(ret)
    return _output_to_list(ret["stdout"])


def plugin_is_enabled(name, runas=None):
    """
    Return whether the plugin is enabled.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.plugin_is_enabled rabbitmq_plugin_name
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    return name in list_enabled_plugins(runas)


def enable_plugin(name, runas=None):
    """
    Enable a RabbitMQ plugin via the rabbitmq-plugins command.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.enable_plugin foo
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    cmd = [_get_rabbitmq_plugin(), "enable", name]
    ret = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    return _format_response(ret, "Enabled")


def disable_plugin(name, runas=None):
    """
    Disable a RabbitMQ plugin via the rabbitmq-plugins command.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.disable_plugin foo
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    cmd = [_get_rabbitmq_plugin(), "disable", name]
    ret = __salt__["cmd.run_all"](
        cmd, reset_system_locale=False, runas=runas, python_shell=False
    )
    return _format_response(ret, "Disabled")


def set_upstream(
    name,
    uri,
    prefetch_count=None,
    reconnect_delay=None,
    ack_mode=None,
    trust_user_id=None,
    exchange=None,
    max_hops=None,
    expires=None,
    message_ttl=None,
    ha_policy=None,
    queue=None,
    runas=None,
):
    """
    Configures an upstream via rabbitmqctl set_parameter. This can be an exchange-upstream,
    a queue-upstream or both.

    :param str name: The name of the upstream to configure.

    The following parameters apply to federated exchanges and federated queues:

    :param str uri: The AMQP URI(s) for the upstream.
    :param int prefetch_count: The maximum number of unacknowledged messages copied
        over a link at any one time. Default: 1000
    :param int reconnect_delay: The duration (in seconds) to wait before reconnecting
        to the broker after being disconnected. Default: 1
    :param str ack_mode: Determines how the link should acknowledge messages.
        If set to ``on-confirm`` (the default), messages are acknowledged to the
        upstream broker after they have been confirmed downstream. This handles
        network errors and broker failures without losing messages, and is the
        slowest option.
        If set to ``on-publish``, messages are acknowledged to the upstream broker
        after they have been published downstream. This handles network errors
        without losing messages, but may lose messages in the event of broker failures.
        If set to ``no-ack``, message acknowledgements are not used. This is the
        fastest option, but may lose messages in the event of network or broker failures.
    :param bool trust_user_id: Determines how federation should interact with the
        validated user-id feature. If set to true, federation will pass through
        any validated user-id from the upstream, even though it cannot validate
        it itself. If set to false or not set, it will clear any validated user-id
        it encounters. You should only set this to true if you trust the upstream
        server (and by extension, all its upstreams) not to forge user-ids.

    The following parameters apply to federated exchanges only:

    :param str exchange: The name of the upstream exchange. Default is to use the
        same name as the federated exchange.
    :param int max_hops: The maximum number of federation links that a message
        published to a federated exchange can traverse before it is discarded.
        Default is 1. Note that even if max-hops is set to a value greater than 1,
        messages will never visit the same node twice due to travelling in a loop.
        However, messages may still be duplicated if it is possible for them to
        travel from the source to the destination via multiple routes.
    :param int expires: The expiry time (in milliseconds) after which an upstream
        queue for a federated exchange may be deleted, if a connection to the upstream
        broker is lost. The default is 'none', meaning the queue should never expire.
        This setting controls how long the upstream queue will last before it is
        eligible for deletion if the connection is lost.
        This value is used to set the "x-expires" argument for the upstream queue.
    :param int message_ttl: The expiry time for messages in the upstream queue
        for a federated exchange (see expires), in milliseconds. Default is ``None``,
        meaning messages should never expire. This does not apply to federated queues.
        This value is used to set the "x-message-ttl" argument for the upstream queue.
    :param str ha_policy: Determines the "x-ha-policy" argument for the upstream
        queue for a federated exchange (see expires). This is only of interest
        when connecting to old brokers which determine queue HA mode using this
        argument. Default is ``None``, meaning the queue is not HA.

    The following parameter applies to federated queues only:

    :param str queue: The name of the upstream queue. Default is to use the same
        name as the federated queue.

    :param str runas: The name of the user to run the command as.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.set_upstream upstream_name ack_mode=on-confirm max_hops=1 \
            trust_user_id=True uri=amqp://hostname

    .. versionadded:: 3000
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    params = salt.utils.data.filter_falsey(
        {
            "uri": uri,
            "prefetch-count": prefetch_count,
            "reconnect-delay": reconnect_delay,
            "ack-mode": ack_mode,
            "trust-user-id": trust_user_id,
            "exchange": exchange,
            "max-hops": max_hops,
            "expires": expires,
            "message-ttl": message_ttl,
            "ha-policy": ha_policy,
            "queue": queue,
        }
    )
    res = __salt__["cmd.run_all"](
        [
            RABBITMQCTL,
            "set_parameter",
            "federation-upstream",
            name,
            salt.utils.json.dumps(params),
        ],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return True


def delete_upstream(name, runas=None):
    """
    Deletes an upstream via rabbitmqctl clear_parameter.

    :param str name: The name of the upstream to delete.
    :param str runas: The name of the user to run the command as.

    CLI Example:

    .. code-block:: bash

        salt '*' rabbitmq.delete_upstream upstream_name

    .. versionadded:: 3000
    """
    if runas is None and not salt.utils.platform.is_windows():
        runas = salt.utils.user.get_user()
    res = __salt__["cmd.run_all"](
        [RABBITMQCTL, "clear_parameter", "federation-upstream", name],
        reset_system_locale=False,
        runas=runas,
        python_shell=False,
    )
    _check_response(res)
    return True



import hashlib
import logging
import os

import salt.auth
from salt.utils.versions import StrictVersion as _StrictVersion

__virtualname__ = os.path.abspath(__file__).rsplit(os.sep)[-2] or "rest_tornado"

log = logging.getLogger(__virtualname__)

# we require at least 4.0, as that includes all the Future's stuff we use
min_tornado_version = "4.0"
has_tornado = False
try:
    import salt.ext.tornado

    if _StrictVersion(salt.ext.tornado.version) >= _StrictVersion(min_tornado_version):
        has_tornado = True
    else:
        log.error("rest_tornado requires at least tornado %s", min_tornado_version)
except (ImportError, TypeError) as err:
    has_tornado = False
    log.error("ImportError! %s", err)


def __virtual__():
    mod_opts = __opts__.get(__virtualname__, {})

    if has_tornado and "port" in mod_opts:
        return __virtualname__

    return False


def get_application(opts):
    try:
        from . import saltnado
    except ImportError as err:
        log.error("ImportError! %s", err)
        return None

    mod_opts = opts.get(__virtualname__, {})

    paths = [
        (r"/", saltnado.SaltAPIHandler),
        (r"/login", saltnado.SaltAuthHandler),
        (r"/minions/(.*)", saltnado.MinionSaltAPIHandler),
        (r"/minions", saltnado.MinionSaltAPIHandler),
        (r"/jobs/(.*)", saltnado.JobsSaltAPIHandler),
        (r"/jobs", saltnado.JobsSaltAPIHandler),
        (r"/run", saltnado.RunSaltAPIHandler),
        (r"/events", saltnado.EventsSaltAPIHandler),
        (r"/hook(/.*)?", saltnado.WebhookSaltAPIHandler),
    ]

    # if you have enabled websockets, add them!
    if mod_opts.get("websockets", False):
        from . import saltnado_websockets

        token_pattern = r"([0-9A-Fa-f]{{{0}}})".format(
            len(getattr(hashlib, opts.get("hash_type", "md5"))().hexdigest())
        )
        all_events_pattern = r"/all_events/{}".format(token_pattern)
        formatted_events_pattern = r"/formatted_events/{}".format(token_pattern)
        log.debug("All events URL pattern is %s", all_events_pattern)
        paths += [
            # Matches /all_events/[0-9A-Fa-f]{n}
            # Where n is the length of hexdigest
            # for the current hashing algorithm.
            # This algorithm is specified in the
            # salt master config file.
            (all_events_pattern, saltnado_websockets.AllEventsHandler),
            (formatted_events_pattern, saltnado_websockets.FormattedEventsHandler),
        ]

    application = salt.ext.tornado.web.Application(
        paths, debug=mod_opts.get("debug", False)
    )

    application.opts = opts
    application.mod_opts = mod_opts
    application.auth = salt.auth.LoadAuth(opts)
    return application


def start():
    """
    Start the saltnado!
    """
    mod_opts = __opts__.get(__virtualname__, {})

    if "num_processes" not in mod_opts:
        mod_opts["num_processes"] = 1

    if mod_opts["num_processes"] > 1 and mod_opts.get("debug", False) is True:
        raise Exception(
            "Tornado's debug implementation is not compatible with multiprocess. "
            "Either disable debug, or set num_processes to 1."
        )

    # the kwargs for the HTTPServer
    kwargs = {}
    if not mod_opts.get("disable_ssl", False):
        if "ssl_crt" not in mod_opts:
            log.error(
                "Not starting '%s'. Options 'ssl_crt' and "
                "'ssl_key' are required if SSL is not disabled.",
                __name__,
            )

            return None
        # cert is required, key may be optional
        # https://docs.python.org/2/library/ssl.html#ssl.wrap_socket
        ssl_opts = {"certfile": mod_opts["ssl_crt"]}
        if mod_opts.get("ssl_key", False):
            ssl_opts.update({"keyfile": mod_opts["ssl_key"]})
        kwargs["ssl_options"] = ssl_opts

    import salt.ext.tornado.httpserver

    http_server = salt.ext.tornado.httpserver.HTTPServer(
        get_application(__opts__), **kwargs
    )
    try:
        http_server.bind(
            mod_opts["port"],
            address=mod_opts.get("address"),
            backlog=mod_opts.get("backlog", 128),
        )
        http_server.start(mod_opts["num_processes"])
    except Exception:  # pylint: disable=broad-except
        log.error(
            "Rest_tornado unable to bind to port %s", mod_opts["port"], exc_info=True
        )
        raise SystemExit(1)

    try:
        salt.ext.tornado.ioloop.IOLoop.current().start()
    except KeyboardInterrupt:
        raise SystemExit(0)



import pytest
import salt.utils.platform
from tests.support.case import SSHCase
from tests.support.unit import skipIf


@skipIf(salt.utils.platform.is_windows(), "salt-ssh not available on Windows")
class SSHRawTest(SSHCase):
    """
    testing salt-ssh with raw calls
    """

    @pytest.mark.slow_test
    def test_ssh_raw(self):
        """
        test salt-ssh with -r argument
        """
        msg = "running raw msg"
        ret = self.run_function("echo {}".format(msg), raw=True)
        self.assertEqual(ret["stdout"], msg + "\n")



"""
Minion status monitoring

Maps to the `status` execution module.
"""

__monitor__ = [
    "loadavg",
    "process",
]


def loadavg(name, maximum=None, minimum=None):
    """
    Return the current load average for the specified minion. Available values
    for name are `1-min`, `5-min` and `15-min`. `minimum` and `maximum` values
    should be passed in as strings.
    """
    # Monitoring state, no changes will be made so no test interface needed
    ret = {
        "name": name,
        "result": False,
        "comment": "",
        "changes": {},
        "data": {},
    }  # Data field for monitoring state

    data = __salt__["status.loadavg"]()
    if name not in data:
        ret["result"] = False
        ret["comment"] += "Requested load average {} not available ".format(name)
        return ret
    if minimum and maximum and minimum >= maximum:
        ret["comment"] += "Min must be less than max"
    if ret["comment"]:
        return ret
    cap = float(data[name])
    ret["data"] = data[name]
    if minimum:
        if cap < float(minimum):
            ret["comment"] = "Load avg is below minimum of {} at {}".format(
                minimum, cap
            )
            return ret
    if maximum:
        if cap > float(maximum):
            ret["comment"] = "Load avg above maximum of {} at {}".format(maximum, cap)
            return ret
    ret["comment"] = "Load avg in acceptable range"
    ret["result"] = True
    return ret


def process(name):
    """
    Return whether the specified signature is found in the process tree. This
    differs slightly from the services states, in that it may refer to a
    process that is not managed via the init system.
    """
    # Monitoring state, no changes will be made so no test interface needed
    ret = {
        "name": name,
        "result": False,
        "comment": "",
        "changes": {},
        "data": {},
    }  # Data field for monitoring state

    data = __salt__["status.pid"](name)
    if not data:
        ret["result"] = False
        ret["comment"] += 'Process signature "{}" not found '.format(name)
        return ret
    ret["data"] = data
    ret["comment"] += 'Process signature "{}" was found '.format(name)
    ret["result"] = True
    return ret



"""
    :codeauthor: jmoney <justin@saltstack.com>
"""


import salt.modules.cp as cp
import salt.transport.client
import salt.utils.files
import salt.utils.platform
import salt.utils.templates as templates
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, mock_open, patch
from tests.support.unit import TestCase


class CpTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.cp module
    """

    def setup_loader_modules(self):
        return {cp: {}}

    def test__render_filenames_undefined_template(self):
        """
        Test if _render_filenames fails upon getting a template not in
        TEMPLATE_REGISTRY.
        """
        path = "/srv/salt/saltines"
        dest = "/srv/salt/cheese"
        saltenv = "base"
        template = "biscuits"
        ret = (path, dest)
        self.assertRaises(
            CommandExecutionError, cp._render_filenames, path, dest, saltenv, template
        )

    def test__render_filenames_render_failed(self):
        """
        Test if _render_filenames fails when template rendering fails.
        """
        path = "salt://saltines"
        dest = "/srv/salt/cheese"
        saltenv = "base"
        template = "jinja"
        file_data = "Remember to keep your files well salted."
        mock_jinja = lambda *args, **kwargs: {"result": False, "data": file_data}
        with patch.dict(templates.TEMPLATE_REGISTRY, {"jinja": mock_jinja}):
            with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
                self.assertRaises(
                    CommandExecutionError,
                    cp._render_filenames,
                    path,
                    dest,
                    saltenv,
                    template,
                )

    def test__render_filenames_success(self):
        """
        Test if _render_filenames succeeds.
        """
        path = "salt://saltines"
        dest = "/srv/salt/cheese"
        saltenv = "base"
        template = "jinja"
        file_data = "/srv/salt/biscuits"
        mock_jinja = lambda *args, **kwargs: {"result": True, "data": file_data}
        ret = (file_data, file_data)  # salt.utils.files.fopen can only be mocked once
        with patch.dict(templates.TEMPLATE_REGISTRY, {"jinja": mock_jinja}):
            with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
                self.assertEqual(
                    cp._render_filenames(path, dest, saltenv, template), ret
                )

    def test_get_file_not_found(self):
        """
        Test if get_file can't find the file.
        """
        with patch("salt.modules.cp.hash_file", MagicMock(return_value=False)):
            path = "salt://saltines"
            dest = "/srv/salt/cheese"
            ret = ""
            self.assertEqual(cp.get_file(path, dest), ret)

    def test_get_file_str_success(self):
        """
        Test if get_file_str succeeds.
        """
        path = "salt://saltines"
        dest = "/srv/salt/cheese/saltines"
        file_data = "Remember to keep your files well salted."
        saltenv = "base"
        ret = file_data
        with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
            with patch("salt.modules.cp.cache_file", MagicMock(return_value=dest)):
                self.assertEqual(cp.get_file_str(path, dest), ret)

    def test_push_non_absolute_path(self):
        """
        Test if push fails on a non absolute path.
        """
        path = "../saltines"
        ret = False

        self.assertEqual(cp.push(path), ret)

    def test_push_dir_non_absolute_path(self):
        """
        Test if push_dir fails on a non absolute path.
        """
        path = "../saltines"
        ret = False

        self.assertEqual(cp.push_dir(path), ret)

    def test_push(self):
        """
        Test if push works with good posix path.
        """
        filename = "/saltines/test.file"
        if salt.utils.platform.is_windows():
            filename = "C:\\saltines\\test.file"
        with patch(
            "salt.modules.cp.os.path",
            MagicMock(isfile=Mock(return_value=True), wraps=cp.os.path),
        ), patch(
            "salt.modules.cp.os.path",
            MagicMock(getsize=MagicMock(return_value=10), wraps=cp.os.path),
        ), patch.multiple(
            "salt.modules.cp",
            _auth=MagicMock(**{"return_value.gen_token.return_value": "token"}),
            __opts__={"id": "abc", "file_buffer_size": 10},
        ), patch(
            "salt.utils.files.fopen", mock_open(read_data=b"content")
        ) as m_open, patch(
            "salt.transport.client.ReqChannel.factory", MagicMock()
        ) as req_channel_factory_mock:
            response = cp.push(filename)
            assert response, response
            num_opens = len(m_open.filehandles[filename])
            assert num_opens == 1, num_opens
            fh_ = m_open.filehandles[filename][0]
            assert fh_.read.call_count == 2, fh_.read.call_count
            req_channel_factory_mock().__enter__().send.assert_called_once_with(
                dict(
                    loc=fh_.tell(),  # pylint: disable=resource-leakage
                    cmd="_file_recv",
                    tok="token",
                    path=["saltines", "test.file"],
                    size=10,
                    data=b"",  # data is empty here because load['data'] is overwritten
                    id="abc",
                )
            )



import os
import tempfile

import pytest
import salt.utils.path
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf


@skipIf(
    salt.utils.path.which_bin(KNOWN_BINARY_NAMES) is None, "virtualenv not installed"
)
class VirtualenvModuleTest(ModuleCase):
    """
    Validate the virtualenv module
    """

    def setUp(self):
        super().setUp()
        self.venv_test_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.venv_dir = os.path.join(self.venv_test_dir, "venv")

    @pytest.mark.slow_test
    def test_create_defaults(self):
        """
        virtualenv.managed
        """
        self.run_function("virtualenv.create", [self.venv_dir])
        pip_file = os.path.join(self.venv_dir, "bin", "pip")
        self.assertTrue(os.path.exists(pip_file))

    @pytest.mark.slow_test
    def test_site_packages(self):
        pip_bin = os.path.join(self.venv_dir, "bin", "pip")
        self.run_function(
            "virtualenv.create", [self.venv_dir], system_site_packages=True
        )
        with_site = self.run_function("pip.freeze", bin_env=pip_bin)
        self.run_function("file.remove", [self.venv_dir])
        self.run_function("virtualenv.create", [self.venv_dir])
        without_site = self.run_function("pip.freeze", bin_env=pip_bin)
        self.assertFalse(with_site == without_site)

    @pytest.mark.slow_test
    def test_clear(self):
        pip_bin = os.path.join(self.venv_dir, "bin", "pip")
        self.run_function("virtualenv.create", [self.venv_dir])
        self.run_function("pip.install", [], pkgs="pep8", bin_env=pip_bin)
        self.run_function("virtualenv.create", [self.venv_dir], clear=True)
        packages = self.run_function("pip.list", prefix="pep8", bin_env=pip_bin)
        self.assertFalse("pep8" in packages)

    def test_virtualenv_ver(self):
        ret = self.run_function("virtualenv.virtualenv_ver", [self.venv_dir])
        assert isinstance(ret, list)
        assert all([isinstance(x, int) for x in ret])

    def tearDown(self):
        self.run_function("file.remove", [self.venv_test_dir])



"""
Module to manage FreeBSD kernel modules
"""

import os
import re

import salt.utils.files

# Define the module's virtual name
__virtualname__ = "kmod"


_LOAD_MODULE = '{0}_load="YES"'
_LOADER_CONF = "/boot/loader.conf"
_MODULE_RE = '^{0}_load="YES"'
_MODULES_RE = r'^(\w+)_load="YES"'


def __virtual__():
    """
    Only runs on FreeBSD systems
    """
    if __grains__["kernel"] == "FreeBSD":
        return __virtualname__
    return (
        False,
        "The freebsdkmod execution module cannot be loaded: only available on FreeBSD"
        " systems.",
    )


def _new_mods(pre_mods, post_mods):
    """
    Return a list of the new modules, pass an kldstat dict before running
    modprobe and one after modprobe has run
    """
    pre = set()
    post = set()
    for mod in pre_mods:
        pre.add(mod["module"])
    for mod in post_mods:
        post.add(mod["module"])
    return post - pre


def _rm_mods(pre_mods, post_mods):
    """
    Return a list of the new modules, pass an kldstat dict before running
    modprobe and one after modprobe has run
    """
    pre = set()
    post = set()
    for mod in pre_mods:
        pre.add(mod["module"])
    for mod in post_mods:
        post.add(mod["module"])
    return pre - post


def _get_module_name(line):
    match = re.search(_MODULES_RE, line)
    if match:
        return match.group(1)
    return None


def _get_persistent_modules():
    """
    Returns a list of modules in loader.conf that load on boot.
    """
    mods = set()
    with salt.utils.files.fopen(_LOADER_CONF, "r") as loader_conf:
        for line in loader_conf:
            line = salt.utils.stringutils.to_unicode(line)
            line = line.strip()
            mod_name = _get_module_name(line)
            if mod_name:
                mods.add(mod_name)
    return mods


def _set_persistent_module(mod):
    """
    Add a module to loader.conf to make it persistent.
    """
    if not mod or mod in mod_list(True) or mod not in available():
        return set()
    __salt__["file.append"](_LOADER_CONF, _LOAD_MODULE.format(mod))
    return {mod}


def _remove_persistent_module(mod, comment):
    """
    Remove module from loader.conf. If comment is true only comment line where
    module is.
    """
    if not mod or mod not in mod_list(True):
        return set()

    if comment:
        __salt__["file.comment"](_LOADER_CONF, _MODULE_RE.format(mod))
    else:
        __salt__["file.sed"](_LOADER_CONF, _MODULE_RE.format(mod), "")

    return {mod}


def available():
    """
    Return a list of all available kernel modules

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.available
    """
    ret = []
    for path in __salt__["file.find"]("/boot/kernel", name="*.ko$"):
        bpath = os.path.basename(path)
        comps = bpath.split(".")
        if "ko" in comps:
            # This is a kernel module, return it without the .ko extension
            ret.append(".".join(comps[: comps.index("ko")]))
    return ret


def check_available(mod):
    """
    Check to see if the specified kernel module is available

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.check_available vmm
    """
    return mod in available()


def lsmod():
    """
    Return a dict containing information about currently loaded modules

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.lsmod
    """
    ret = []
    for line in __salt__["cmd.run"]("kldstat").splitlines():
        comps = line.split()
        if not len(comps) > 2:
            continue
        if comps[0] == "Id":
            continue
        if comps[4] == "kernel":
            continue
        ret.append({"module": comps[4][:-3], "size": comps[3], "depcount": comps[1]})
    return ret


def mod_list(only_persist=False):
    """
    Return a list of the loaded module names

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.mod_list
    """
    mods = set()
    if only_persist:
        if not _get_persistent_modules():
            return mods
        for mod in _get_persistent_modules():
            mods.add(mod)
    else:
        for mod in lsmod():
            mods.add(mod["module"])
    return sorted(list(mods))


def load(mod, persist=False):
    """
    Load the specified kernel module

    mod
        Name of the module to add

    persist
        Write the module to sysrc kld_modules to make it load on system reboot

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.load bhyve
    """
    pre_mods = lsmod()
    response = __salt__["cmd.run_all"]("kldload {}".format(mod), python_shell=False)
    if response["retcode"] == 0:
        post_mods = lsmod()
        mods = _new_mods(pre_mods, post_mods)
        persist_mods = set()
        if persist:
            persist_mods = _set_persistent_module(mod)
        return sorted(list(mods | persist_mods))
    elif "module already loaded or in kernel" in response["stderr"]:
        if persist and mod not in _get_persistent_modules():
            persist_mods = _set_persistent_module(mod)
            return sorted(list(persist_mods))
        else:
            # It's compiled into the kernel
            return [None]
    else:
        return "Module {} not found".format(mod)


def is_loaded(mod):
    """
    Check to see if the specified kernel module is loaded

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.is_loaded vmm
    """
    return mod in mod_list()


def remove(mod, persist=False, comment=True):
    """
    Remove the specified kernel module

    mod
        Name of module to remove

    persist
        Also remove module from /boot/loader.conf

    comment
        If persist is set don't remove line from /boot/loader.conf but only
        comment it

    CLI Example:

    .. code-block:: bash

        salt '*' kmod.remove vmm
    """
    pre_mods = lsmod()
    res = __salt__["cmd.run_all"]("kldunload {}".format(mod), python_shell=False)
    if res["retcode"] == 0:
        post_mods = lsmod()
        mods = _rm_mods(pre_mods, post_mods)
        persist_mods = set()
        if persist:
            persist_mods = _remove_persistent_module(mod, comment)
        return sorted(list(mods | persist_mods))
    else:
        return "Error removing module {}: {}".format(mod, res["stderr"])



"""
    :codeauthor: Andrew Colin Kissa <andrew@topdog.za.net>
"""

import pytest
import salt.states.postgres_privileges as postgres_privileges
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {postgres_privileges: {}}


def test_present_table():
    """
    Test present
    """
    table_name = "awl"
    name = "baruwa"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        postgres_privileges.__salt__, {"postgres.has_privileges": mock_true}
    ):
        comt = "The requested privilege(s) are already set"
        ret.update({"comment": comt, "result": True})
        assert postgres_privileges.present(name, table_name, "table") == ret

    with patch.dict(
        postgres_privileges.__salt__,
        {
            "postgres.has_privileges": mock_false,
            "postgres.privileges_grant": mock_true,
        },
    ):
        with patch.dict(postgres_privileges.__opts__, {"test": True}):
            comt = "The privilege(s): {} are set to be granted to {}".format(
                "ALL", name
            )
            ret.update({"comment": comt, "result": None})
            assert (
                postgres_privileges.present(
                    name, table_name, "table", privileges=["ALL"]
                )
                == ret
            )

        with patch.dict(postgres_privileges.__opts__, {"test": False}):
            comt = "The privilege(s): {} have been granted to {}".format("ALL", name)
            ret.update(
                {"comment": comt, "result": True, "changes": {"baruwa": "Present"}}
            )
            assert (
                postgres_privileges.present(
                    name, table_name, "table", privileges=["ALL"]
                )
                == ret
            )


def test_present_group():
    """
    Test present group
    """
    group_name = "admins"
    name = "baruwa"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        postgres_privileges.__salt__,
        {
            "postgres.has_privileges": mock_false,
            "postgres.privileges_grant": mock_true,
        },
    ):
        with patch.dict(postgres_privileges.__opts__, {"test": True}):
            comt = "The privilege(s): {} are set to be granted to {}".format(
                group_name, name
            )
            ret.update({"comment": comt, "result": None})
            assert postgres_privileges.present(name, group_name, "group") == ret

        with patch.dict(postgres_privileges.__opts__, {"test": False}):
            comt = "The privilege(s): {} have been granted to {}".format(
                group_name, name
            )
            ret.update(
                {"comment": comt, "result": True, "changes": {"baruwa": "Present"}}
            )
            assert postgres_privileges.present(name, group_name, "group") == ret


def test_absent_table():
    """
    Test absent
    """
    table_name = "awl"
    name = "baruwa"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        postgres_privileges.__salt__, {"postgres.has_privileges": mock_false}
    ):
        with patch.dict(postgres_privileges.__opts__, {"test": True}):
            comt = "The requested privilege(s) are not set so cannot be revoked"
            ret.update({"comment": comt, "result": True})
            assert postgres_privileges.absent(name, table_name, "table") == ret

    with patch.dict(
        postgres_privileges.__salt__,
        {
            "postgres.has_privileges": mock_true,
            "postgres.privileges_revoke": mock_true,
        },
    ):
        with patch.dict(postgres_privileges.__opts__, {"test": True}):
            comt = "The privilege(s): {} are set to be revoked from {}".format(
                "ALL", name
            )
            ret.update({"comment": comt, "result": None})
            assert (
                postgres_privileges.absent(
                    name, table_name, "table", privileges=["ALL"]
                )
                == ret
            )

        with patch.dict(postgres_privileges.__opts__, {"test": False}):
            comt = "The privilege(s): {} have been revoked from {}".format("ALL", name)
            ret.update(
                {"comment": comt, "result": True, "changes": {"baruwa": "Absent"}}
            )
            assert (
                postgres_privileges.absent(
                    name, table_name, "table", privileges=["ALL"]
                )
                == ret
            )


def test_absent_group():
    """
    Test absent group
    """
    group_name = "admins"
    name = "baruwa"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    with patch.dict(
        postgres_privileges.__salt__,
        {
            "postgres.has_privileges": mock_true,
            "postgres.privileges_revoke": mock_true,
        },
    ):
        with patch.dict(postgres_privileges.__opts__, {"test": True}):
            comt = "The privilege(s): {} are set to be revoked from {}".format(
                group_name, name
            )
            ret.update({"comment": comt, "result": None})
            assert postgres_privileges.absent(name, group_name, "group") == ret

        with patch.dict(postgres_privileges.__opts__, {"test": False}):
            comt = "The privilege(s): {} have been revoked from {}".format(
                group_name, name
            )
            ret.update(
                {"comment": comt, "result": True, "changes": {"baruwa": "Absent"}}
            )
            assert postgres_privileges.absent(name, group_name, "group") == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.rbenv as rbenv
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rbenv: {}}


def test_installed():
    """
    Test to verify that the specified ruby is installed with rbenv.
    """
    # rbenv.is_installed is used wherever test is False.
    mock_is = MagicMock(side_effect=[False, True, True, True, True])

    # rbenv.install is only called when an action is attempted
    # (ie. Successfully... or Failed...)
    mock_i = MagicMock(side_effect=[False, False, False])

    # rbenv.install_ruby is only called when rbenv is successfully
    # installed and an attempt to install a version of Ruby is
    # made.
    mock_ir = MagicMock(side_effect=[True, False])
    mock_def = MagicMock(return_value="2.3.4")
    mock_ver = MagicMock(return_value=["2.3.4", "2.4.1"])
    with patch.dict(
        rbenv.__salt__,
        {
            "rbenv.is_installed": mock_is,
            "rbenv.install": mock_i,
            "rbenv.default": mock_def,
            "rbenv.versions": mock_ver,
            "rbenv.install_ruby": mock_ir,
        },
    ):
        with patch.dict(rbenv.__opts__, {"test": True}):
            name = "1.9.3-p551"
            comt = "Ruby {} is set to be installed".format(name)
            ret = {"name": name, "changes": {}, "comment": comt, "result": None}
            assert rbenv.installed(name) == ret

            name = "2.4.1"
            comt = "Ruby {} is already installed".format(name)
            ret = {
                "name": name,
                "changes": {},
                "comment": comt,
                "default": False,
                "result": True,
            }
            assert rbenv.installed(name) == ret

            name = "2.3.4"
            comt = "Ruby {} is already installed".format(name)
            ret = {
                "name": name,
                "changes": {},
                "comment": comt,
                "default": True,
                "result": True,
            }
            assert rbenv.installed(name) == ret

        with patch.dict(rbenv.__opts__, {"test": False}):
            name = "2.4.1"
            comt = "Rbenv failed to install"
            ret = {"name": name, "changes": {}, "comment": comt, "result": False}
            assert rbenv.installed(name) == ret

            comt = "Requested ruby exists"
            ret = {
                "name": name,
                "comment": comt,
                "default": False,
                "changes": {},
                "result": True,
            }
            assert rbenv.installed(name) == ret

            name = "2.3.4"
            comt = "Requested ruby exists"
            ret = {
                "name": name,
                "comment": comt,
                "default": True,
                "changes": {},
                "result": True,
            }
            assert rbenv.installed(name) == ret

            name = "1.9.3-p551"
            comt = "Successfully installed ruby"
            ret = {
                "name": name,
                "comment": comt,
                "default": False,
                "changes": {name: "Installed"},
                "result": True,
            }
            assert rbenv.installed(name) == ret

            comt = "Failed to install ruby"
            ret = {"name": name, "comment": comt, "changes": {}, "result": False}
            assert rbenv.installed(name) == ret


def test_absent():
    """
    Test to verify that the specified ruby is not installed with rbenv.
    """
    # rbenv.is_installed is used for all tests here.
    mock_is = MagicMock(
        side_effect=[False, True, True, True, False, True, True, True, True, True]
    )
    # rbenv.uninstall_ruby is only called when an action is
    # attempted (ie. Successfully... or Failed...)
    mock_uninstalled = MagicMock(side_effect=[True, False, False, True])
    mock_def = MagicMock(return_value="2.3.4")
    mock_ver = MagicMock(return_value=["2.3.4", "2.4.1"])
    with patch.dict(
        rbenv.__salt__,
        {
            "rbenv.is_installed": mock_is,
            "rbenv.default": mock_def,
            "rbenv.versions": mock_ver,
            "rbenv.uninstall_ruby": mock_uninstalled,
        },
    ):

        with patch.dict(rbenv.__opts__, {"test": True}):
            name = "1.9.3-p551"
            comt = "Rbenv not installed, {} not either".format(name)
            ret = {"name": name, "changes": {}, "comment": comt, "result": True}
            assert rbenv.absent(name) == ret

            comt = "Ruby {} is already uninstalled".format(name)
            ret = {"name": name, "changes": {}, "comment": comt, "result": True}
            assert rbenv.absent(name) == ret

            name = "2.3.4"
            comt = "Ruby {} is set to be uninstalled".format(name)
            ret = {
                "name": name,
                "changes": {},
                "comment": comt,
                "default": True,
                "result": None,
            }
            assert rbenv.absent("2.3.4") == ret

            name = "2.4.1"
            comt = "Ruby {} is set to be uninstalled".format(name)
            ret = {
                "name": name,
                "changes": {},
                "comment": comt,
                "default": False,
                "result": None,
            }
            assert rbenv.absent("2.4.1") == ret

        with patch.dict(rbenv.__opts__, {"test": False}):
            name = "1.9.3-p551"
            comt = "Rbenv not installed, {} not either".format(name)
            ret = {"name": name, "changes": {}, "comment": comt, "result": True}
            assert rbenv.absent(name) == ret

            comt = "Ruby {} is already absent".format(name)
            ret = {"name": name, "changes": {}, "comment": comt, "result": True}
            assert rbenv.absent(name) == ret

            name = "2.3.4"
            comt = "Successfully removed ruby"
            ret = {
                "name": name,
                "changes": {name: "Uninstalled"},
                "comment": comt,
                "default": True,
                "result": True,
            }
            assert rbenv.absent(name) == ret

            comt = "Failed to uninstall ruby"
            ret = {
                "name": name,
                "changes": {},
                "comment": comt,
                "default": True,
                "result": False,
            }
            assert rbenv.absent(name) == ret

            name = "2.4.1"
            comt = "Failed to uninstall ruby"
            ret = {
                "name": name,
                "changes": {},
                "comment": comt,
                "default": False,
                "result": False,
            }
            assert rbenv.absent(name) == ret

            comt = "Successfully removed ruby"
            ret = {
                "name": name,
                "changes": {name: "Uninstalled"},
                "comment": comt,
                "default": False,
                "result": True,
            }
            assert rbenv.absent(name) == ret


def test_install_rbenv():
    """
    Test to install rbenv if not installed.
    """
    name = "myqueue"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock_is = MagicMock(side_effect=[False, True, True, False, False])
    mock_i = MagicMock(side_effect=[False, True])
    with patch.dict(
        rbenv.__salt__, {"rbenv.is_installed": mock_is, "rbenv.install": mock_i}
    ):

        with patch.dict(rbenv.__opts__, {"test": True}):
            comt = "Rbenv is set to be installed"
            ret.update({"comment": comt, "result": None})
            assert rbenv.install_rbenv(name) == ret

            comt = "Rbenv is already installed"
            ret.update({"comment": comt, "result": True})
            assert rbenv.install_rbenv(name) == ret

        with patch.dict(rbenv.__opts__, {"test": False}):
            comt = "Rbenv is already installed"
            ret.update({"comment": comt, "result": True})
            assert rbenv.install_rbenv(name) == ret

            comt = "Rbenv failed to install"
            ret.update({"comment": comt, "result": False})
            assert rbenv.install_rbenv(name) == ret

            comt = "Rbenv installed"
            ret.update({"comment": comt, "result": True})
            assert rbenv.install_rbenv(name) == ret
